"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import sys
from collections import defaultdict
from hub.exceptions import ModuleNotInstalledException, OutOfBoundsError
from hub.schema.features import Primitive, Tensor, SchemaDict
from hub.schema import Audio, BBox, ClassLabel, Image, Sequence, Text, Video
from .dataset import Dataset
import hub.store.pickle_s3_storage
import hub.schema.serialize
import hub.schema.deserialize


def _to_pytorch(
    dataset,
    transform=None,
    inplace=True,
    output_type=dict,
    indexes=None,
):
    """| Converts the dataset into a pytorch compatible format.

    Parameters
    ----------
    transform: function that transforms data in a dict format
    inplace: bool, optional
        Defines if data should be converted to torch.Tensor before or after Transforms applied (depends on what data
        type you need for Transforms). Default is True.
    output_type: one of list, tuple, dict, optional
        Defines the output type. Default is dict - same as in original Hub Dataset.
    indexes: list or int, optional
        The samples to be converted into tensorflow format. Takes all samples in dataset by default.
    """
    try:
        import torch
    except ModuleNotFoundError:
        raise ModuleNotInstalledException("torch")

    global torch
    indexes = indexes or dataset.indexes

    if "r" not in dataset.mode:
        dataset.flush()  # FIXME Without this some tests in test_converters.py fails, not clear why
    return TorchDataset(
        dataset, transform, inplace=inplace, output_type=output_type, indexes=indexes
    )


def _from_pytorch(dataset, scheduler: str = "single", workers: int = 1):
    """| Converts a pytorch dataset object into hub format

    Parameters
    ----------
    dataset:
        The pytorch dataset object that needs to be converted into hub format
    scheduler: str
        choice between "single", "threaded", "processed"
    workers: int
        how many threads or processes to use
    """

    if "torch" not in sys.modules:
        raise ModuleNotInstalledException("torch")
    else:
        import torch

        global torch

    max_dict = defaultdict(lambda: None)

    def sampling(ds):
        for sample in ds:
            dict_sampling(sample)

    def dict_sampling(d, path=""):
        for k, v in d.items():
            k = k.replace("/", "_")
            cur_path = path + "/" + k
            if isinstance(v, dict):
                dict_sampling(v, path=cur_path)
            elif isinstance(v, str):
                if cur_path not in max_dict.keys():
                    max_dict[cur_path] = (len(v),)
                else:
                    max_dict[cur_path] = max(((len(v)),), max_dict[cur_path])
            elif hasattr(v, "shape"):
                if cur_path not in max_dict.keys():
                    max_dict[cur_path] = v.shape
                else:
                    max_dict[cur_path] = tuple(
                        [max(value) for value in zip(max_dict[cur_path], v.shape)]
                    )

    sampling(dataset)

    def generate_schema(dataset):
        sample = dataset[0]
        return dict_to_hub(sample).dict_

    def dict_to_hub(dic, path=""):
        d = {}
        for k, v in dic.items():
            k = k.replace("/", "_")
            cur_path = path + "/" + k
            if isinstance(v, dict):
                d[k] = dict_to_hub(v, path=cur_path)
            else:
                value_shape = v.shape if hasattr(v, "shape") else ()
                if isinstance(v, torch.Tensor):
                    v = v.numpy()
                shape = tuple(None for it in value_shape)
                max_shape = (
                    max_dict[cur_path] or tuple(10000 for it in value_shape)
                    if not isinstance(v, str)
                    else (10000,)
                )
                dtype = v.dtype.name if hasattr(v, "dtype") else type(v)
                dtype = "int64" if isinstance(v, str) else dtype
                d[k] = (
                    Tensor(shape=shape, dtype=dtype, max_shape=max_shape)
                    if not isinstance(v, str)
                    else Text(shape=(None,), dtype=dtype, max_shape=max_shape)
                )
        return SchemaDict(d)

    my_schema = generate_schema(dataset)

    def transform_numpy(sample):
        d = {}
        for k, v in sample.items():
            k = k.replace("/", "_")
            d[k] = transform_numpy(v) if isinstance(v, dict) else v
        return d

    @hub.transform(schema=my_schema, scheduler=scheduler, workers=workers)
    def my_transform(sample):
        return transform_numpy(sample)

    return my_transform(dataset)


def _to_tensorflow(dataset, indexes=None, include_shapes=False, key_list=None):
    """| Converts the dataset into a tensorflow compatible format

    Parameters
    ----------
    indexes: list or int, optional
        The samples to be converted into tensorflow format. Takes all samples in dataset by default.
    include_shapes: boolean, optional
        False by default. Setting it to True passes the shapes to tf.data.Dataset.from_generator.
        Setting to True could lead to issues with dictionaries inside Tensors.
    """
    try:
        import tensorflow as tf

        global tf
    except ModuleNotFoundError:
        raise ModuleNotInstalledException("tensorflow")
    key_list = key_list or list(dataset.keys)
    key_list = [key if key.startswith("/") else "/" + key for key in key_list]
    for key in key_list:
        if key not in dataset.keys:
            raise KeyError()
    indexes = indexes or dataset.indexes
    indexes = [indexes] if isinstance(indexes, int) else indexes
    _samples_in_chunks = {
        key: value.chunks[0] for key, value in dataset._tensors.items()
    }
    _active_chunks = {}
    _active_chunks_range = {}

    def _get_active_item(key, index):
        active_range = _active_chunks_range.get(key)
        samples_per_chunk = _samples_in_chunks[key]
        if active_range is None or index not in active_range:
            active_range_start = index - index % samples_per_chunk
            active_range = range(
                active_range_start,
                min(active_range_start + samples_per_chunk, indexes[-1] + 1),
            )
            _active_chunks_range[key] = active_range
            _active_chunks[key] = dataset._tensors[key][
                active_range.start : active_range.stop
            ]
        return _active_chunks[key][index % samples_per_chunk]

    def tf_gen():
        key_dtype_map = {key: dataset[key, indexes[0]].dtype for key in dataset.keys}
        for index in indexes:
            d = {}
            for key in dataset.keys:
                if key not in key_list:
                    continue
                split_key, cur = key.split("/"), d
                for i in range(1, len(split_key) - 1):
                    if split_key[i] in cur.keys():
                        cur = cur[split_key[i]]
                    else:
                        cur[split_key[i]] = {}
                        cur = cur[split_key[i]]
                cur[split_key[-1]] = _get_active_item(key, index)
                if isinstance(key_dtype_map[key], Text):
                    value = cur[split_key[-1]]
                    cur[split_key[-1]] = (
                        "".join(chr(it) for it in value.tolist())
                        if value.ndim == 1
                        else ["".join(chr(it) for it in val.tolist()) for val in value]
                    )

            yield (d)

    def dict_to_tf(my_dtype, path=""):
        d = {}
        for k, v in my_dtype.dict_.items():
            for key in key_list:
                if key.startswith(path + "/" + k):
                    d[k] = dtype_to_tf(v, path + "/" + k)
                    break
        return d

    def tensor_to_tf(my_dtype):
        return dtype_to_tf(my_dtype.dtype)

    def text_to_tf(my_dtype):
        return "string"

    def dtype_to_tf(my_dtype, path=""):
        if isinstance(my_dtype, SchemaDict):
            return dict_to_tf(my_dtype, path=path)
        elif isinstance(my_dtype, Text):
            return text_to_tf(my_dtype)
        elif isinstance(my_dtype, Tensor):
            return tensor_to_tf(my_dtype)
        elif isinstance(my_dtype, Primitive):
            if str(my_dtype._dtype) == "object":
                return "string"
            return str(my_dtype._dtype)

    def get_output_shapes(my_dtype, path=""):
        if isinstance(my_dtype, SchemaDict):
            return output_shapes_from_dict(my_dtype, path=path)
        elif isinstance(my_dtype, (Text, Primitive)):
            return ()
        elif isinstance(my_dtype, Tensor):
            return my_dtype.shape

    def output_shapes_from_dict(my_dtype, path=""):
        d = {}
        for k, v in my_dtype.dict_.items():
            for key in key_list:
                if key.startswith(path + "/" + k):
                    d[k] = get_output_shapes(v, path + "/" + k)
                    break
        return d

    output_types = dtype_to_tf(dataset._schema)
    if include_shapes:
        output_shapes = get_output_shapes(dataset._schema)
        return tf.data.Dataset.from_generator(
            tf_gen, output_types=output_types, output_shapes=output_shapes
        )
    else:
        return tf.data.Dataset.from_generator(tf_gen, output_types=output_types)


def _from_tensorflow(ds, scheduler: str = "single", workers: int = 1):
    """Converts a tensorflow dataset into hub format.

    Parameters
    ----------
    dataset:
        The tensorflow dataset object that needs to be converted into hub format
    scheduler: str
        choice between "single", "threaded", "processed"
    workers: int
        how many threads or processes to use

    Examples
    --------
    >>> ds = tf.data.Dataset.from_tensor_slices(tf.range(10))
    >>> out_ds = hub.Dataset.from_tensorflow(ds)
    >>> res_ds = out_ds.store("username/new_dataset") # res_ds is now a usable hub dataset

    >>> ds = tf.data.Dataset.from_tensor_slices({'a': [1, 2], 'b': [5, 6]})
    >>> out_ds = hub.Dataset.from_tensorflow(ds)
    >>> res_ds = out_ds.store("username/new_dataset") # res_ds is now a usable hub dataset

    >>> ds = hub.Dataset(schema=my_schema, shape=(1000,), url="username/dataset_name", mode="w")
    >>> ds = ds.to_tensorflow()
    >>> out_ds = hub.Dataset.from_tensorflow(ds)
    >>> res_ds = out_ds.store("username/new_dataset") # res_ds is now a usable hub dataset
    """
    if "tensorflow" not in sys.modules:
        raise ModuleNotInstalledException("tensorflow")
    else:
        import tensorflow as tf

        global tf

    def generate_schema(ds):
        if isinstance(ds._structure, tf.TensorSpec):
            return tf_to_hub({"data": ds._structure}).dict_
        return tf_to_hub(ds._structure).dict_

    def tf_to_hub(tf_dt):
        if isinstance(tf_dt, dict):
            return dict_to_hub(tf_dt)
        elif isinstance(tf_dt, tf.TensorSpec):
            return TensorSpec_to_hub(tf_dt)

    def TensorSpec_to_hub(tf_dt):
        dt = tf_dt.dtype.name if tf_dt.dtype.name != "string" else "object"
        shape = tuple(tf_dt.shape) if tf_dt.shape.rank is not None else (None,)
        return Tensor(shape=shape, dtype=dt)

    def dict_to_hub(tf_dt):
        d = {key.replace("/", "_"): tf_to_hub(value) for key, value in tf_dt.items()}
        return SchemaDict(d)

    my_schema = generate_schema(ds)

    def transform_numpy(sample):
        d = {}
        for k, v in sample.items():
            k = k.replace("/", "_")
            if not isinstance(v, dict):
                if isinstance(v, (tuple, list)):
                    new_v = list(v)
                    for i in range(len(new_v)):
                        new_v[i] = new_v[i].numpy()
                    d[k] = tuple(new_v) if isinstance(v, tuple) else new_v
                else:
                    d[k] = v.numpy()
            else:
                d[k] = transform_numpy(v)
        return d

    @hub.transform(schema=my_schema, scheduler=scheduler, workers=workers)
    def my_transform(sample):
        sample = sample if isinstance(sample, dict) else {"data": sample}
        return transform_numpy(sample)

    return my_transform(ds)


def _from_tfds(
    dataset,
    split=None,
    num: int = -1,
    sampling_amount: int = 1,
    scheduler: str = "single",
    workers: int = 1,
):
    """| Converts a TFDS Dataset into hub format.

    Parameters
    ----------
    dataset: str
        The name of the tfds dataset that needs to be converted into hub format
    split: str, optional
        A string representing the splits of the dataset that are required such as "train" or "test+train"
        If not present, all the splits of the dataset are used.
    num: int, optional
        The number of samples required. If not present, all the samples are taken.
        If count is -1, or if count is greater than the size of this dataset, the new dataset will contain all elements of this dataset.
    sampling_amount: float, optional
        a value from 0 to 1, that specifies how much of the dataset would be sampled to determinte feature shapes
        value of 0 would mean no sampling and 1 would imply that entire dataset would be sampled
    scheduler: str
        choice between "single", "threaded", "processed"
    workers: int
        how many threads or processes to use

    Examples
    --------
    >>> out_ds = hub.Dataset.from_tfds('mnist', split='test+train', num=1000)
    >>> res_ds = out_ds.store("username/mnist") # res_ds is now a usable hub dataset
    """
    try:
        import tensorflow_datasets as tfds

        global tfds
    except Exception:
        raise ModuleNotInstalledException("tensorflow_datasets")

    ds_info = tfds.load(dataset, with_info=True)

    if split is None:
        all_splits = ds_info[1].splits.keys()
        split = "+".join(all_splits)

    ds = tfds.load(dataset, split=split)
    ds = ds.take(num)
    max_dict = defaultdict(lambda: None)

    def sampling(ds):
        try:
            subset_len = len(ds) if hasattr(ds, "__len__") else num
        except Exception:
            subset_len = max(num, 5)

        subset_len = int(max(subset_len * sampling_amount, 5))
        samples = ds.take(subset_len)
        for smp in samples:
            dict_sampling(smp)

    def dict_sampling(d, path=""):
        for k, v in d.items():
            k = k.replace("/", "_")
            cur_path = path + "/" + k
            if isinstance(v, dict):
                dict_sampling(v)
            elif hasattr(v, "shape") and v.dtype != "string":
                if cur_path not in max_dict.keys():
                    max_dict[cur_path] = v.shape
                else:
                    max_dict[cur_path] = tuple(
                        [max(value) for value in zip(max_dict[cur_path], v.shape)]
                    )
            elif hasattr(v, "shape") and v.dtype == "string":
                if cur_path not in max_dict.keys():
                    max_dict[cur_path] = (len(v.numpy()),)
                else:
                    max_dict[cur_path] = max(((len(v.numpy()),), max_dict[cur_path]))

    if sampling_amount > 0:
        sampling(ds)

    def generate_schema(ds):
        tf_schema = ds[1].features
        return to_hub(tf_schema).dict_

    def to_hub(tf_dt, max_shape=None, path=""):
        if isinstance(tf_dt, tfds.features.FeaturesDict):
            return sdict_to_hub(tf_dt, path=path)
        elif isinstance(tf_dt, tfds.features.Image):
            return image_to_hub(tf_dt, max_shape=max_shape)
        elif isinstance(tf_dt, tfds.features.ClassLabel):
            return class_label_to_hub(tf_dt, max_shape=max_shape)
        elif isinstance(tf_dt, tfds.features.Video):
            return video_to_hub(tf_dt, max_shape=max_shape)
        elif isinstance(tf_dt, tfds.features.Text):
            return text_to_hub(tf_dt, max_shape=max_shape)
        elif isinstance(tf_dt, tfds.features.Sequence):
            return sequence_to_hub(tf_dt, max_shape=max_shape)
        elif isinstance(tf_dt, tfds.features.BBoxFeature):
            return bbox_to_hub(tf_dt, max_shape=max_shape)
        elif isinstance(tf_dt, tfds.features.Audio):
            return audio_to_hub(tf_dt, max_shape=max_shape)
        elif isinstance(tf_dt, tfds.features.Tensor):
            return tensor_to_hub(tf_dt, max_shape=max_shape)
        else:
            if tf_dt.dtype.name != "string":
                return tf_dt.dtype.name

    def sdict_to_hub(tf_dt, path=""):
        d = {}
        for key, value in tf_dt.items():
            key = key.replace("/", "_")
            cur_path = path + "/" + key
            d[key] = to_hub(value, max_dict[cur_path], cur_path)
        return SchemaDict(d)

    def tensor_to_hub(tf_dt, max_shape=None):
        if tf_dt.dtype.name == "string":
            max_shape = max_shape or (100000,)
            return Text(shape=(None,), dtype="int64", max_shape=(100000,))
        dt = tf_dt.dtype.name
        if max_shape and len(max_shape) > len(tf_dt.shape):
            max_shape = max_shape[(len(max_shape) - len(tf_dt.shape)) :]

        max_shape = max_shape or tuple(
            10000 if dim is None else dim for dim in tf_dt.shape
        )
        return Tensor(shape=tf_dt.shape, dtype=dt, max_shape=max_shape)

    def image_to_hub(tf_dt, max_shape=None):
        dt = tf_dt.dtype.name
        if max_shape and len(max_shape) > len(tf_dt.shape):
            max_shape = max_shape[(len(max_shape) - len(tf_dt.shape)) :]

        max_shape = max_shape or tuple(
            10000 if dim is None else dim for dim in tf_dt.shape
        )
        return Image(
            shape=tf_dt.shape,
            dtype=dt,
            max_shape=max_shape,  # compressor="png"
        )

    def class_label_to_hub(tf_dt, max_shape=None):
        if hasattr(tf_dt, "_num_classes"):
            return ClassLabel(
                num_classes=tf_dt.num_classes,
            )
        else:
            return ClassLabel(names=tf_dt.names)

    def text_to_hub(tf_dt, max_shape=None):
        max_shape = max_shape or (100000,)
        dt = "int64"
        return Text(shape=(None,), dtype=dt, max_shape=max_shape)

    def bbox_to_hub(tf_dt, max_shape=None):
        dt = tf_dt.dtype.name
        return BBox(dtype=dt)

    def sequence_to_hub(tf_dt, max_shape=None):
        return Sequence(dtype=to_hub(tf_dt._feature), shape=())

    def audio_to_hub(tf_dt, max_shape=None):
        if max_shape and len(max_shape) > len(tf_dt.shape):
            max_shape = max_shape[(len(max_shape) - len(tf_dt.shape)) :]

        max_shape = max_shape or tuple(
            100000 if dim is None else dim for dim in tf_dt.shape
        )
        dt = tf_dt.dtype.name
        return Audio(
            shape=tf_dt.shape,
            dtype=dt,
            max_shape=max_shape,
            file_format=tf_dt._file_format,
            sample_rate=tf_dt._sample_rate,
        )

    def video_to_hub(tf_dt, max_shape=None):
        if max_shape and len(max_shape) > len(tf_dt.shape):
            max_shape = max_shape[(len(max_shape) - len(tf_dt.shape)) :]

        max_shape = max_shape or tuple(
            10000 if dim is None else dim for dim in tf_dt.shape
        )
        dt = tf_dt.dtype.name
        return Video(shape=tf_dt.shape, dtype=dt, max_shape=max_shape)

    my_schema = generate_schema(ds_info)

    def transform_numpy(sample):
        d = {}
        for k, v in sample.items():
            k = k.replace("/", "_")
            d[k] = transform_numpy(v) if isinstance(v, dict) else v.numpy()
        return d

    @hub.transform(schema=my_schema, scheduler=scheduler, workers=workers)
    def my_transform(sample):
        return transform_numpy(sample)

    return my_transform(ds)


class TorchDataset:
    def __init__(
        self, ds, transform=None, inplace=True, output_type=dict, indexes=None
    ):
        self._ds = None
        self._url = ds.url
        self._token = ds.token
        self._transform = transform
        self.inplace = inplace
        self.output_type = output_type
        self.indexes = indexes
        self._inited = False

    def _do_transform(self, data):
        return self._transform(data) if self._transform else data

    def _init_ds(self):
        """
        For each process, dataset should be independently loaded
        """
        if self._ds is None:
            self._ds = Dataset(self._url, token=self._token, lock_cache=False)
        if not self._inited:
            self._inited = True
            self._samples_in_chunks = {
                key: (None in value.shape) and 1 or value.chunks[0]
                for key, value in self._ds._tensors.items()
            }
            self._active_chunks = {}
            self._active_chunks_range = {}

    def __len__(self):
        self._init_ds()
        return len(self.indexes) if isinstance(self.indexes, list) else 1

    def _get_active_item(self, key, index):
        active_range = self._active_chunks_range.get(key)
        samples_per_chunk = self._samples_in_chunks[key]
        if active_range is None or index not in active_range:
            active_range_start = index - index % samples_per_chunk
            active_range = range(
                active_range_start, active_range_start + samples_per_chunk
            )
            self._active_chunks_range[key] = active_range
            self._active_chunks[key] = self._ds._tensors[key][
                active_range.start : active_range.stop
            ]
        return self._active_chunks[key][index % samples_per_chunk]

    def __getitem__(self, ind):
        if isinstance(self.indexes, int):
            if ind != 0:
                raise OutOfBoundsError(f"Got index {ind} for dataset of length 1")
            index = self.indexes
        else:
            index = self.indexes[ind]
        self._init_ds()
        d = {}
        for key in self._ds._tensors.keys():
            split_key = key.split("/")
            cur = d
            for i in range(1, len(split_key) - 1):
                if split_key[i] not in cur.keys():
                    cur[split_key[i]] = {}
                cur = cur[split_key[i]]

            item = self._get_active_item(key, index)
            if not isinstance(item, bytes) and not isinstance(item, str):
                t = item
                if self.inplace:
                    t = torch.tensor(t)
                cur[split_key[-1]] = t
        d = self._do_transform(d)
        if self.inplace & (self.output_type != dict) & (isinstance(d, dict)):
            d = self.output_type(d.values())
        return d

    def __iter__(self):
        self._init_ds()
        for i in range(len(self)):
            yield self[i]



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from collections.abc import MutableMapping

import posixpath


class NestedStore(MutableMapping):
    def __init__(self, storage: MutableMapping, root: str):
        self._storage = storage
        self._root = posixpath.normpath(root)

    def __getitem__(self, k):
        return self._storage[posixpath.join(self._root, k)]

    def __setitem__(self, k, v):
        self._storage[posixpath.join(self._root, k)] = v

    def __delitem__(self, k):
        del self._storage[posixpath.join(self._root, k)]

    def __iter__(self):
        prefix = self._root + "/"
        for item in self._storage:
            item: str
            if item.startswith(prefix):
                yield item[len(prefix) :]

    def __len__(self):
        return sum(1 for _ in self)

    def flush(self):
        self._storage.flush()

    def commit(self):
        """ Deprecated alias to flush()"""
        self.flush()

    def close(self):
        self._storage.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from typing import Tuple

from hub.schema.features import Tensor


class Audio(Tensor):

    """Schema for audio would define the maximum shape of the audio dataset and their respective sampling rate.

    Example: This example uploads an `audio file` to a Hub dataset `audio_dataset` with `HubSchema` and retrieves it.

    ----------
    >>> import hub
    >>> from hub.schema import Audio
    >>> from hub import transform, schema
    >>> import librosa
    >>> from librosa import display
    >>> import numpy as np

    >>> # Define schema
    >>> my_schema={
    >>>     "wav": Audio(shape=(None,), max_shape=(1920000,), file_format="wav", dtype=float),
    >>>     "sampling_rate": Primitive(dtype=int),
    >>> }
    >>>
    >>> sample = glob("audio.wav")

    >>> # Define transform
    >>> @transform(schema=my_schema)
    >>> def load_transform(sample):
    >>>     audio, sr = librosa.load(sample, sr=None)
    >>>
    >>>     return {
    >>>         "wav": audio,
    >>>         "sampling_rate": sr
    >>>     }
    >>>
    >>> # Returns a transform object
    >>> ds = load_transform(sample)

    >>> # Load data
    >>> ds = Dataset(tag)
    >>>
    >>> tag = "username/audio_dataset"
    >>>
    >>> # Pushes to hub
    >>> ds2 = ds.store(tag)

    >>> # Fetching from Hub
    >>> data = Dataset(tag)
    >>>
    >>> # Fetch the first sample
    >>> audio_sample = data["wav"][0].compute()
    >>>
    >>> # Audio file
        array([ 9.15527344e-05,  2.13623047e-04,  0.00000000e+00, ...,
        -2.73132324e-02, -2.99072266e-02, -2.44750977e-02])


    """

    def __init__(
        self,
        shape: Tuple[int, ...] = (None,),
        dtype="int64",
        file_format=None,
        sample_rate: int = None,
        max_shape: Tuple[int, ...] = None,
        chunks=None,
        compressor="lz4",
    ):
        """Constructs the connector.

        Parameters
        ----------
        file_format: `str`
            the audio file format. Can be any format ffmpeg
            understands. If `None`, will attempt to infer from the file extension.
        shape: `tuple`
             shape of the data.
        dtype: str
            The dtype of the data.
        sample_rate: `int`
            additional metadata exposed to the user through
            `info.schema['audio'].sample_rate`. This value isn't used neither in
            encoding nor decoding.


        Raises
        ----------
        ValueError: If the shape is invalid
        """
        self.file_format = file_format
        if len(shape) != 1:
            raise ValueError(
                f"Audio schema currently only supports 1-D values, got {shape}"
            )
        # self._shape = shape
        self.sample_rate = sample_rate
        super().__init__(
            shape=shape,
            dtype=dtype,
            max_shape=max_shape,
            chunks=chunks,
            compressor=compressor,
        )

    def __str__(self):
        out = super().__str__()
        out = "Audio" + out[6:-1]
        out = (
            out + ", file_format=" + self.file_format
            if self.file_format is not None
            else out
        )
        out = (
            out + ", sample_rate=" + str(self.sample_rate)
            if self.sample_rate is not None
            else out
        )
        out += ")"
        return out

    def __repr__(self):
        return self.__str__()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from typing import Tuple, Dict, Iterable
import hub

Shape = Tuple[int, ...]


class FlatTensor:
    """ Tensor metadata after applying flatten function """

    def __init__(self, path: str, shape: Shape, dtype, max_shape: Shape, chunks: Shape):
        self.path = path
        self.shape = shape
        self.dtype = dtype
        self.max_shape = max_shape
        self.chunks = chunks


class HubSchema:
    """ Base class for all datatypes"""

    def _flatten(self) -> Iterable[FlatTensor]:
        """ Flattens dtype into list of tensors that will need to be stored seperately """
        raise NotImplementedError()


class Primitive(HubSchema):
    """Class for handling primitive datatypes.
    All numpy primitive data types like int32, float64, etc... should be wrapped around this class.
    """

    def __init__(self, dtype, chunks=None, compressor="lz4"):
        self._dtype = hub.dtype(dtype)
        self.chunks = _normalize_chunks(chunks)
        self.shape = self.max_shape = ()
        self.dtype = self._dtype
        self.compressor = compressor

    def _flatten(self):
        yield FlatTensor("", (), self._dtype, (), self.chunks)

    def __str__(self):
        return "'" + str(self.dtype) + "'"

    def __repr__(self):
        return self.__str__()


class SchemaDict(HubSchema):
    """Class for dict branching of a datatype.
    SchemaDict dtype contains str -> dtype associations.
    This way you can describe complex datatypes.
    """

    def __init__(self, dict_):
        self.dict_: Dict[str, HubSchema] = {
            key: featurify(value) for key, value in dict_.items()
        }

    def _flatten(self):
        for key, value in self.dict_.items():
            for item in value._flatten():
                yield FlatTensor(
                    f"/{key}{item.path}",
                    item.shape,
                    item.dtype,
                    item.max_shape,
                    item.chunks,
                )

    def __str__(self):
        out = "SchemaDict("
        out += str(self.dict_)
        out += ")"
        return out

    def __repr__(self):
        return self.__str__()


def featurify(schema) -> HubSchema:
    """This functions converts naked primitive datatypes and ditcs into Primitives and SchemaDicts.
    That way every node in dtype tree is a SchemaConnector type object.
    """
    if isinstance(schema, dict):
        return SchemaDict(schema)
    elif isinstance(schema, HubSchema):
        return schema
    else:
        return Primitive(schema)


def _normalize_chunks(chunks):
    chunks = (chunks,) if isinstance(chunks, int) else chunks
    chunks = tuple(chunks) if chunks else None
    return chunks


class Tensor(HubSchema):
    """Tensor type in schema.
    Has np-array like structure contains any type of elements (Primitive and non-Primitive).
    Tensors can't be visualized at app.activeloop.ai.
    """

    def __init__(
        self,
        shape: Shape = (None,),
        dtype="float64",
        max_shape: Shape = None,
        chunks=None,
        compressor="lz4",
    ):
        """
        Parameters
        ----------
        shape : Tuple[int]
            Shape of tensor, can contains None(s) meaning the shape can be dynamic
            Dynamic shape means it can change during editing the dataset
        dtype : SchemaConnector or str
            dtype of each element in Tensor. Can be Primitive and non-Primitive type
        max_shape : Tuple[int]
            Maximum shape of tensor shape if tensor is dynamic
        chunks : Tuple[int] | True
            Describes how to split tensor dimensions into chunks (files) to store them efficiently.
            It is anticipated that each file should be ~16MB.
            Sample Count is also in the list of tensor's dimensions (first dimension)
            If default value is chosen, automatically detects how to split into chunks
        """
        if shape is None:
            raise TypeError("shape cannot be None")
        if isinstance(shape, Iterable) and None in shape and max_shape is None:
            raise ValueError(
                "while specifying shape containing None dimensions, max_shape argument needs to be provided"
            )
        if not isinstance(shape, (tuple, int, list)):
            raise TypeError(f"shape of {type(shape)} is not supported")

        shape = (shape,) if isinstance(shape, int) else tuple(shape)
        for dim in shape:
            if not isinstance(dim, int) and dim is not None:
                raise TypeError(f"shape can't have {type(dim)} in its dimension")
        max_shape = shape if max_shape is None else max_shape
        if not isinstance(max_shape, (tuple, int, list)):
            raise TypeError(f"max_shape of {type(max_shape)} is not supported")
        max_shape = (max_shape,) if isinstance(max_shape, int) else tuple(max_shape)
        for dim in max_shape:
            if dim is None:
                raise TypeError("max_shape can't have None in it's dimension")
            elif not isinstance(dim, int):
                raise TypeError(f"max_shape can't have {type(dim)} in its dimension")

        if len(shape) != len(max_shape):
            raise ValueError(
                f"shape {shape} and max_shape {max_shape} have different lengths"
            )
        for dim, max_dim in zip(shape, max_shape):
            if dim is not None and dim != max_dim:
                raise ValueError(f"shape and max_shape mismatch, {dim} != {max_dim}")

        chunks = _normalize_chunks(chunks)

        # TODO add errors if shape and max_shape have wrong values
        self.shape = tuple(shape)
        self.dtype = featurify(dtype)
        self.max_shape = max_shape
        self.chunks = chunks
        self.compressor = compressor

    def _flatten(self):
        for item in self.dtype._flatten():
            yield FlatTensor(
                item.path,
                self.shape + item.shape,
                item.dtype,
                self.max_shape + item.max_shape,
                self.chunks or item.chunks,
            )

    def __str__(self):
        out = "Tensor(shape=" + str(self.shape) + ", dtype=" + str(self.dtype)
        out = (
            out + ", max_shape=" + str(self.max_shape)
            if self.max_shape != self.shape
            else out
        )
        out = out + ", chunks=" + str(self.chunks) if self.chunks is not None else out
        out += ")"
        return out

    def __repr__(self):
        return self.__str__()


def flatten(dtype, root=""):
    """ Flattens nested dictionary and returns tuple (dtype, path) """
    dtype = featurify(dtype)
    if isinstance(dtype, SchemaDict):
        for key, value in dtype.dict_.items():
            yield from flatten(value, root + "/" + key)
    else:
        yield (dtype, root)



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import io
import time
import random
import traceback
import json

import numpy as np
from PIL import Image as PILImage

from .base import Base


class Image(Base):
    def __init__(self, format: str):
        self._format = format
        super().__init__()

    def encode(self, array: np.ndarray) -> bytes:
        raise NotImplementedError("Not ready yet")
        info = {"shape": array.shape, "dtype": str(array.dtype)}

        assert array.dtype == "uint8"
        assert len(array.shape) >= 3
        assert array.shape[-1] == 3

        if len(array.shape) == 3:
            with io.BytesIO() as bs:
                img = PILImage.fromarray(array)
                img.save(bs, format=self._format)
                return bs.getvalue()
        else:
            images = []
            for i, index in enumerate(np.ndindex(array.shape[:-3])):
                bs = io.BytesIO()
                img = PILImage.fromarray(array[index])
                img.save(bs, format=self._format)
                images.append(bs.getvalue())
        info["images"] = images
        return msgpack.dumps(info)

    def decode(self, content: bytes) -> np.ndarray:
        raise NotImplementedError("Not ready yet")
        info = msgpack.loads(content)
        if "image" in info:
            img = PILImage.open(io.BytesIO(bytearray(info["image"])))
            img.reshape(info["shape"])
            return img
        else:
            array = np.zeros(info["shape"], info["dtype"])
            images = info["images"]
            for i, index in enumerate(np.ndindex(info["shape"][:-3])):
                img = PILImage.open(io.BytesIO(bytearray(images[i])))
                arr = np.asarray(img)
                array[index] = arr
            return array



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

# NOT IMPLEMENTED COMPLETELY YET!

from typing import Union
from enum import Enum

from hub.defaults import AZURE_HOST_SUFFIX


class UrlProtocol(Enum):
    UNKNOWN = "unknown"
    S3 = "s3"
    GCS = "gcs"
    AZURE = "azure"
    FILESYSTEM = "filesystem"


class UrlType(Enum):
    HUB = "hub"
    LOCAL = "local"
    CLOUD = "cloud"


class Url:
    @classmethod
    def parse(cls, url: str) -> "Url":
        assert isinstance(url, str)

        pass

    def __init__(
        self,
        url_type: UrlType,
        protocol: UrlProtocol,
        path: str,  # for get_mapper(path)
        bucket: Union[str, None] = None,
        user: Union[str, None] = None,
        dataset: Union[str, None] = None,
        endpoint_url: Union[str, None] = None,
    ):
        assert isinstance(url_type, UrlType)
        assert isinstance(protocol, UrlProtocol)
        assert isinstance(path, str)
        assert isinstance(bucket, str) or bucket is None
        assert isinstance(user, str) or user is None
        assert isinstance(dataset, str) or dataset is None
        assert isinstance(endpoint_url, str) or endpoint_url is None

        self.url_type = url_type
        self.protocol = protocol
        self.path = path
        self.bucket = bucket
        self.user = user
        self.dataset = dataset
        self.endpoint_url = endpoint_url

    @property
    def url(self) -> str:
        pass



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""
from collections import defaultdict
import json
from collections.abc import MutableMapping
import posixpath
from hub import defaults


# TODO: Better version control for PB scale data
# Retrieving chunks with version control is fast, but the memory used to store the chunk_commit_map would be huge ~GBs
class MetaStorage(MutableMapping):
    @classmethod
    def to_str(cls, obj):
        if isinstance(obj, memoryview):
            obj = obj.tobytes()
        if isinstance(obj, bytes):
            obj = obj.decode("utf-8")
        return obj

    def __init__(self, path, fs_map: MutableMapping, meta_map: MutableMapping, ds):
        self._fs_map = fs_map
        self._meta = meta_map
        self._path = path
        self._ds = ds

    def find_chunk(self, k: str) -> str:
        ls = self._ds._chunk_commit_map[self._path][k]
        cur_node = self._ds._version_node
        while cur_node is not None:
            if cur_node.commit_id in ls:
                return f"{k}:{cur_node.commit_id}"
            cur_node = cur_node.parent
        return None

    def __getitem__(self, k: str, check=True) -> bytes:
        filename = posixpath.split(k)[1]
        if filename.startswith("."):
            return bytes(
                json.dumps(
                    json.loads(self.to_str(self._meta[defaults.META_FILE]))[k][
                        self._path
                    ]
                ),
                "utf-8",
            )
        if check:
            if self._ds._commit_id:
                k = self.find_chunk(k) or f"{k}:{self._ds._commit_id}"
        return self._fs_map[k]

    def get(self, k: str, check=True) -> bytes:
        filename = posixpath.split(k)[1]
        if filename.startswith("."):
            meta_ = self._meta.get(defaults.META_FILE)
            if not meta_:
                return None
            meta = json.loads(self.to_str(meta_))
            metak = meta.get(k)
            if not metak:
                return None
            item = metak.get(self._path)
            return bytes(json.dumps(item), "utf-8") if item else None
        else:
            if check:
                if self._ds._commit_id:
                    k = self.find_chunk(k) or f"{k}:{self._ds._commit_id}"
            return self._fs_map.get(k)

    def __setitem__(self, k: str, v: bytes, check=True):
        filename = posixpath.split(k)[1]
        if filename.startswith("."):
            meta = json.loads(self.to_str(self._meta[defaults.META_FILE]))
            meta[k] = meta.get(k) or {}
            meta[k][self._path] = json.loads(self.to_str(v))
            self._meta[defaults.META_FILE] = bytes(json.dumps(meta), "utf-8")
        else:
            chunk_key = k.split(":")[0]
            if check and self._ds._commit_id:
                old_filename = self.find_chunk(k)
                k = f"{k}:{self._ds._commit_id}"
                if old_filename and k != old_filename:
                    self.copy_chunk(old_filename, k)
            commit_id = k.split(":")[-1]
            self._ds._chunk_commit_map[self._path][chunk_key].add(commit_id)
            self._fs_map[k] = v

    def copy_all_chunks(self, from_commit_id: str, to_commit_id: str):
        ls = {
            chunk
            for chunk, commit_ids in self._ds._chunk_commit_map[self._path].items()
            if from_commit_id in commit_ids
        }

        for chunk in ls:
            from_path = f"{chunk}:{from_commit_id}"
            to_path = f"{chunk}:{to_commit_id}"
            self.copy_chunk(from_path, to_path)

    def copy_chunk(self, from_chunk: str, to_chunk: str):
        data = self.__getitem__(from_chunk, False)
        self.__setitem__(to_chunk, data, False)

    def __len__(self):
        return len(self._fs_map) + 1

    def __iter__(self):
        yield ".zarray"
        yield from self._fs_map

    def __delitem__(self, k: str):
        filename = posixpath.split(k)[1]
        if not filename.startswith("."):
            filename = self.find_chunk(filename) or f"{filename}:{self._ds._commit_id}"
        if filename.startswith("."):
            meta = json.loads(self.to_str(self._meta[defaults.META_FILE]))
            meta[k] = meta.get(k) or dict()
            meta[k][self._path] = None
            self._meta[defaults.META_FILE] = bytes(json.dumps(meta), "utf-8")
        else:
            chunk_key = k.split(":")[0]
            if self._ds._commit_id:
                k = self.find_chunk(k) or f"{k}:{self._ds._commit_id}"
            commit_id = k.split(":")[-1]
            self._ds._chunk_commit_map[self._path][chunk_key].remove(commit_id)
            del self._fs_map[k]

    def flush(self):
        self._meta.flush()
        self._fs_map.flush()

    def commit(self):
        """ Deprecated alias to flush()"""
        self.flush()

    def close(self):
        self._meta.close()
        self._fs_map.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from typing import MutableMapping, Tuple
import posixpath
import shutil
import configparser
import os
from time import sleep

import re
import fsspec
import gcsfs
import zarr

from hub.store.lru_cache import LRUCache
from hub.client.hub_control import HubControlClient
from hub.store.azure_fs import AzureBlobFileSystem
from hub.store.s3_file_system_replacement import S3FileSystemReplacement


def get_user_name():
    creds = HubControlClient().get_config()
    return creds["_id"]


def _connect(tag, public=True):
    """Connects to the backend and receives credentials"""

    creds = HubControlClient().get_config()
    dataset = HubControlClient().get_dataset_path(tag)

    # If dataset is in DB then return the path
    # Otherwise construct the path from the tag
    if dataset and "path" in dataset:
        path = dataset["path"]
    else:
        sub_tags = tag.split("/")
        # Get repository path from the cred location
        path = "/".join(creds["bucket"].split("/")[:-2])
        path = path + "/public" if public else path + "/private"
        path = f"{path}/{sub_tags[0]}/{sub_tags[-1]}"
    return path, creds


def get_fs_and_path(
    url: str, token=None, public=True
) -> Tuple[fsspec.AbstractFileSystem, str]:
    if url.startswith("s3://"):
        token = token or dict()
        token = read_aws_creds(token) if isinstance(token, str) else token
        return (
            S3FileSystemReplacement(
                key=token.get("aws_access_key_id"),
                secret=token.get("aws_secret_access_key"),
                token=token.get("aws_session_token"),
                client_kwargs={
                    "endpoint_url": token.get("endpoint_url"),
                    "region_name": token.get("region"),
                },
            ),
            url[5:],
        )
    elif url.startswith("gcs://"):
        return gcsfs.GCSFileSystem(token=token), url[6:]
    elif url.find("blob.core.windows.net/") != -1:
        account_name = url.split(".")[0]
        account_name = account_name[8:] if url.startswith("https://") else account_name
        return (
            AzureBlobFileSystem(
                account_name=account_name,
                account_key=token.get("account_key"),
            ),
            url[url.find("blob.core.windows.net/") + 22 :],
        )
    elif (
        url.startswith("../")
        or url.startswith("./")
        or url.startswith("/")
        or url.startswith("~/")
    ):
        return fsspec.filesystem("file"), url
    elif (
        # windows local file system
        re.search("^[A-Za-z]:", url)
    ):
        return fsspec.filesystem("file"), url
    else:
        # TOOD check if url is username/dataset:version
        url, creds = _connect(url, public=public)
        fs = S3FileSystemReplacement(
            key=creds["access_key"],
            secret=creds["secret_key"],
            token=creds["session_token"],
            client_kwargs={
                "endpoint_url": creds["endpoint"],
                "region_name": creds["region"],
            },
        )
        return (fs, url)


def read_aws_creds(filepath: str):
    parser = configparser.ConfigParser()
    parser.read(filepath)
    return {section: dict(parser.items(section)) for section in parser.sections()}


def _get_storage_map(fs, path):
    return StorageMapWrapperWithCommit(fs.get_mapper(path, check=False, create=False))


def get_cache_path(path, cache_folder="~/.activeloop/cache/"):
    if path.startswith("s3://") or path.startswith("gcs://"):
        path = "//".join(path.split("//")[1:])
    elif (
        path.startswith("../")
        or path.startswith("./")
        or path.startswith("/")
        or path.startswith("~/")
    ):
        path = "/".join(path.split("/")[1:])
    elif path.find("://") != -1:
        path = path.split("://")[-1]
    elif path.find(":\\") != -1:
        path = path.split(":\\")[-1]
    else:
        # path is username/dataset or username/dataset:version
        path = path.replace(":", "/")
    return os.path.expanduser(posixpath.join(cache_folder, path))


def get_storage_map(fs, path, memcache=2 ** 26, lock=True, storage_cache=2 ** 28):
    store = _get_storage_map(fs, path)
    if memcache and memcache > 0:
        store = LRUCache(zarr.MemoryStore(), store, memcache)
    return store


class StorageMapWrapperWithCommit(MutableMapping):
    def __init__(self, map):
        self._map = map
        self.root = self._map.root

    def __getitem__(self, slice_):
        return self._map[slice_]

    def __setitem__(self, slice_, value):
        self._map[slice_] = value

    def __delitem__(self, slice_):
        del self._map[slice_]

    def __len__(self):
        return len(self._map)

    def __iter__(self):
        yield from self._map

    def flush(self):
        pass

    def commit(self):
        """ Deprecated alias to flush()"""
        self.flush()

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from collections import abc
from configparser import ConfigParser
import json
import os
from typing import Dict, Tuple
import sys

import fsspec
import numpy as np
import traceback
from gcsfs.core import GCSFileSystem

from hub.client.hub_control import HubControlClient
from hub.codec import Base as BaseCodec
from hub.codec import from_name as codec_from_name
from hub.collections.tensor.core import Tensor
from hub.collections.client_manager import get_client
from hub.log import logger
from hub.exceptions import (
    PermissionException,
    HubDatasetNotFoundException,
    ModuleNotInstalledException,
)

from hub.utils import _flatten


class Transform:
    def __init__(self):
        pass

    def __call__(self, input):
        return self.forward(input)

    def meta(self):
        """
        Provides the metadata for all tensors including shapes, dtypes, dtags and chunksize for each array in the form

        Returns
        -------
        returns
            dict of tensor

        Examples
        -------
        >>> def meta()
        >>>     return {
        >>>         ...
        >>>         "tesnor_name":{
        >>>             "shape": (1,256,256),
        >>>             "dtype": "uint8",
        >>>             "chunksize": 100,
        >>>             "dtag": "segmentation"
        >>>         }
        >>>         ...
        >>>     }
        """

        raise NotImplementedError()

    def forward(input):
        """
        Takes a an element of a list or sample from dataset and returns sample of the dataset

        Parameters
        -------
        input
            an element of list or dict of arrays

        Returns
        -------
        dict
            dict of numpy arrays

        Examples
        -------
        >>> def forward(input):
        >>>    ds = {}
        >>>    ds["image"] = np.empty(1, object)
        >>>    ds["image"][0] = np.array(256, 256)
        >>>    return ds
        """
        raise NotImplementedError()


DatasetGenerator = Transform


def _numpy_to_tuple(arr: np.ndarray):
    """Converts numpy array to tuple of numpy arrays"""
    return [np.array([t]) for t in arr]


def _numpy_saver(
    fs: fsspec.AbstractFileSystem, filepath: str, array: np.ndarray, codec: BaseCodec
):
    """Saves a single numpy array into filepath given specific filesystem"""
    with fs.open(filepath, "wb") as f:
        f.write(codec.encode(array))


def _numpy_saver_multi(
    fs: fsspec.AbstractFileSystem, filepath: str, arrays: np.ndarray, offset: int
):
    for i in range(len(arrays)):
        _numpy_saver(fs, f"{filepath}/{offset+i}.npy", arrays[i : i + 1])
    return len(arrays)


def _preprocess_meta_before_save(meta: dict):
    meta = dict(meta)
    meta["dtype"] = str(meta["dtype"])
    return meta


def _dask_shape(input_shape: Tuple[int]):
    """Dask accept np.nan value in shape if the axis length is not known, our API uses -1 for that, this function converts -1 to np.nan"""
    return (np.nan,) + input_shape[1:] if input_shape[0] == -1 else input_shape


def _dict_to_tuple(d: dict):
    """Converts dict of lists into (flattened list of values, list of keys)"""
    keys = sorted(d.keys())
    lens = {len(d[key]) for key in keys}
    assert len(lens) == 1
    cnt = next(iter(lens))
    return [d[key][i] for i in range(cnt) for key in keys], keys


def _tuple_to_dict(t: tuple, keys: tuple):
    """Converts (flattened list of values, list of keys) into dict of lists"""
    cnt = len(keys)
    assert len(t) % cnt == 0
    return {key: [t[i] for i in range(j, len(t), cnt)] for j, key in enumerate(keys)}


def _load_creds(creds):
    """Loads credentials from "{creds}" cfg file if such exists
    if creds is dict, then dict will be returned, assuming all credential data is in dict
    """
    if creds is None:
        return None
    elif isinstance(creds, str) and os.path.exists(creds):
        parser = ConfigParser()
        parser.read(creds)
        return {section: dict(parser.items(section)) for section in parser.sections()}
    else:
        return creds


def _connect(tag):
    """Connects to the backend and receive credentials"""

    creds = HubControlClient().get_config()
    dataset = HubControlClient().get_dataset_path(tag)

    if dataset and "path" in dataset:
        path = dataset["path"]
    else:
        sub_tags = tag.split("/")
        real_tag = sub_tags[-1]
        if len(sub_tags) > 1 and sub_tags[0] != creds["_id"]:
            username = creds["bucket"].split("/")[-1]
            creds["bucket"] = creds["bucket"].replace(username, sub_tags[0])

        path = f"{creds['bucket']}/{real_tag}"
    return path, creds


def _load_fs_and_path(path, creds=None, session_creds=True, google_cloud_project=""):
    """Given url(path) and creds returns filesystem required for accessing that file + url's filepath in that filesystem"""
    if (
        path.startswith("./")
        or path.startswith("/")
        or path.startswith("../")
        or path.startswith("~/")
    ):
        return fsspec.filesystem("file"), os.path.expanduser(path.replace("fs://", ""))

    if (
        session_creds
        and creds is None
        and not path.startswith("s3://")
        and not path.startswith("gcs://")
    ):
        path, creds = _connect(path)

    if path.startswith("s3://"):
        path = path[5:]
        if creds is not None and session_creds:

            return (
                fsspec.filesystem(
                    "s3",
                    key=creds["access_key"],
                    secret=creds["secret_key"],
                    token=creds["session_token"],
                    client_kwargs={
                        "endpoint_url": creds["endpoint"],
                        "region_name": creds["region"],
                    },
                ),
                path,
            )
        elif creds is not None:
            return (
                fsspec.filesystem(
                    "s3",
                    key=creds.get("access_key"),
                    secret=creds.get("secret_key"),
                ),
                path,
            )
        else:
            return fsspec.filesystem("s3"), path
    elif path.startswith("gcs://"):
        return (
            GCSFileSystem(project=google_cloud_project, token=creds),
            path[6:],
        )


class Dataset:
    def __init__(self, tensors: Dict[str, Tensor], metainfo=dict()):
        """Creates dict given dict of tensors (name -> Tensor key value pairs)"""
        self._tensors = tensors
        self._metainfo = metainfo
        shape = None
        for name, tensor in tensors.items():
            if shape is None or tensor.ndim > len(shape):
                shape = tensor.shape
            self._len = tensor.count
        self.verison = "0.x"
        if "dask" not in sys.modules:
            raise ModuleNotInstalledException("dask")
        else:
            import dask
            import dask.array

            global dask

    def __len__(self) -> int:
        """len of dataset (len of tensors across axis 0, yes, they all should be = to each other)
        Raises Exception if length is unknown
        """
        if self._len == -1:
            raise Exception(
                "Cannot return __len__ of dataset for which __len__ is not known, use .count property, it will return -1 instead of this Exception"
            )
        return self._len

    @property
    def license(self) -> str:
        """Dataset license"""
        return self._metainfo.get("license") if self._metainfo else None

    @property
    def description(self) -> str:
        """Dataset description"""
        return self._metainfo.get("description") if self._metainfo else None

    @property
    def citation(self) -> str:
        """Dataset citation"""
        return self._metainfo.get("citation") if self._metainfo else None

    @property
    def howtoload(self) -> str:
        """Dataset howtoload"""
        return self._metainfo.get("howtoload") if self._metainfo else None

    @property
    def count(self) -> int:
        """len of dataset (len of tensors across axis 0, yes, they all should be = to each other)
        Returns -1 if length is unknown
        """
        return self._len

    def __iter__(self):
        """Iterates over axis 0 return dict of Tensors"""
        for i in range(len(self)):
            yield {key: t._array[i] for key, t in self._tensors.items()}

    def keys(self):
        """Returns names of tensors"""
        yield from self._tensors.keys()

    def values(self):
        """Returns tensors"""
        yield from self._tensors.values()

    def items(self):
        """Returns tensors"""
        yield from self._tensors.items()

    def __getitem__(self, slices) -> "Dataset":
        """Returns a slice of dataset
        slices can be
            1) List of strs (slicing horizontally)
            2) List of slices or ints (slicing vertically)
            3) Both (1) and (2) at the same time
            4) Single int, slice, str is also accepted
        """
        if isinstance(slices, tuple):
            if all([isinstance(s, str) for s in slices]):
                return Dataset({key: self._tensors[key] for key in slices})
            elif isinstance(slices[0], abc.Iterable) and all(
                [isinstance(s, str) for s in slices[0]]
            ):
                return Dataset({key: self._tensors[key] for key in slices[0]})[
                    slices[1:]
                ]
            else:
                assert all(
                    [isinstance(s, slice) or isinstance(s, int) for s in slices]
                ), "invalid indexing, either wrong order or wrong type"
                ndim = len(slices)
                if all(isinstance(s, int) for s in slices):
                    return {
                        name: tensor[slices]
                        for name, tensor in self._tensors.items()
                        if tensor.ndim >= ndim
                    }
                else:
                    return Dataset(
                        {
                            name: tensor[slices]
                            for name, tensor in self._tensors.items()
                            if tensor.ndim >= ndim
                        }
                    )

        elif isinstance(slices, str):
            return self._tensors[slices]
        elif isinstance(slices, slice):
            return Dataset({key: value[slices] for key, value in self._tensors.items()})
        elif isinstance(slices, int):
            return {key: value[slices] for key, value in self._tensors.items()}

    def cache(self) -> "Dataset":
        raise NotImplementedError()

    def _store_unknown_sized_ds(self, fs: fsspec.AbstractFileSystem, path: str) -> int:
        client = get_client()
        worker_count = sum(client.ncores().values())
        # worker_count = 1
        chunks = {key: t._delayed_objs for key, t in self._tensors.items()}
        chunk_count = [len(items) for _, items in chunks.items()]
        assert (
            len(set(chunk_count)) == 1
        ), "Number of chunks in each tensor should be the same to be able to store dataset"
        chunk_count = chunk_count[0]
        count = 0
        collected = {el: None for el in self._tensors.keys()}
        collected_offset = {el: 0 for el in collected}
        # max_chunksize = max(*[t.chunksize for t in self._tensors])
        for i in range(0, chunk_count, worker_count):
            batch_count = min(i + worker_count, chunk_count) - i
            lasttime = True if i + worker_count >= chunk_count else False
            tasks = {
                key: delayed_objs[i : i + batch_count]
                for key, delayed_objs in chunks.items()
            }
            # logger.info(tasks)
            tasks, keys = _dict_to_tuple(tasks)

            # dask.visualize(
            #     tasks, filename=f"./data/tasks/{i}", optimize_graph=True,
            # )
            persisted = client.persist(tasks)
            persisted = _tuple_to_dict(persisted, keys)
            # for j in range(batch_count):
            #     assert (
            #         len(
            #             {
            #                 # len(objs[j])
            #                 # client.submit()
            #                 dask.delayed(len)(objs[j]).compute()
            #                 for objs in persisted.values()
            #             }
            #         )
            #         == 1
            #     ), "All numpy arrays returned from call should have same len"
            lens = {
                key: [dask.delayed(len)(objs[j]) for j in range(batch_count)]
                for key, objs in persisted.items()
            }
            lens, keys = _dict_to_tuple(lens)
            lens = client.gather(client.compute(lens))
            lens = _tuple_to_dict(lens, keys)
            for key, objs in persisted.items():
                arr = _dask_concat(
                    [
                        dask.array.from_delayed(
                            obj,
                            dtype=self._tensors[key].dtype,
                            shape=(lens[key][i],) + tuple(self._tensors[key].shape[1:]),
                        )
                        for i, obj in enumerate(objs)
                    ]
                )
                if collected[key] is None:
                    collected[key] = arr
                else:
                    collected[key] = _dask_concat([collected[key], arr])
            # tasks = [obj for key, objs in persisted.items() for obj in objs]
            tasks = []

            for key in list(collected.keys()):
                c = collected[key]
                chunksize = self._tensors[key].chunksize
                codec = codec_from_name(self._tensors[key].dcompress)
                cnt = len(c) - len(c) % chunksize if not lasttime else len(c)
                for i in range(0, cnt, chunksize):
                    tasks += [
                        dask.delayed(_numpy_saver)(
                            fs,
                            f"{path}/{key}/{collected_offset[key] + i}.npy",
                            c[i : i + chunksize],
                            codec,
                        )
                    ]
                collected_offset[key] += cnt
                collected[key] = collected[key][cnt:]
            client.gather(client.compute(tasks))
        count = set(collected_offset.values())
        assert (
            len(count) == 1
        ), "All tensors should be the same size to be stored in the same dataset"
        return next(iter(count))

    def _store_known_sized_ds(self, fs: fsspec.AbstractFileSystem, path: str) -> int:
        client = get_client()
        worker_count = sum(client.ncores().values())
        # chunksize = min(*[t.chunksize for t in self._tensors.values()])
        chunksize = (
            min(*[t.chunksize for t in self._tensors.values()])
            if len(self._tensors) > 1
            else next(iter(self._tensors.values())).chunksize
        )
        cnt = len(self)
        collected = {el: None for el in self._tensors.keys()}
        collected_offset = {el: 0 for el in collected}
        step = worker_count * chunksize
        for i in range(0, cnt, step):
            batch_count = min(step, cnt - i)
            lasttime = True if i + step >= cnt else False
            persisted = client.persist(
                [self._tensors[key]._array[i : i + batch_count] for key in collected]
            )
            persisted = {key: persisted[j] for j, key in enumerate(collected)}
            tasks = []
            for el, arr in persisted.items():
                if collected[el] is None:
                    collected[el] = arr
                else:
                    collected[el] = _dask_concat([collected[el], arr])
                c = collected[el]
                chunksize_ = self._tensors[el].chunksize
                codec = codec_from_name(self._tensors[el].dcompress)
                if len(c) >= chunksize_ or lasttime:
                    jcnt = len(c) - len(c) % chunksize_ if not lasttime else len(c)
                    for j in range(0, jcnt, chunksize_):
                        tasks += [
                            dask.delayed(_numpy_saver)(
                                fs,
                                f"{path}/{el}/{collected_offset[el] + j}.npy",
                                collected[el][j : j + chunksize_],
                                codec,
                            )
                        ]
                    collected_offset[el] += jcnt
                    collected[el] = collected[el][jcnt:]
            client.gather(client.compute(tasks))
        count = set(collected_offset.values())
        assert (
            len(count) == 1
        ), "All tensors should be the same size to be stored in the same dataset"
        return next(iter(count))

    @property
    def meta(self) -> dict:
        """Dict of meta's of each tensor
        meta of tensor contains all metadata for tensor storage
        """
        tensor_meta = {
            name: _preprocess_meta_before_save(t._meta)
            for name, t in self._tensors.items()
        }
        ds_meta = {"tensors": tensor_meta, "len": self.count}
        return ds_meta

    def delete(self, tag, creds=None, session_creds=True) -> bool:
        """Deletes dataset given tag(filepath) and credentials (optional)"""
        fs, path = _load_fs_and_path(tag, creds, session_creds=session_creds)
        fs: fsspec.AbstractFileSystem = fs
        if fs.exists(path):
            fs.delete(path, recursive=True)
            return True
        return False

    def store(self, tag, creds=None, session_creds=True) -> "Dataset":
        """Stores dataset by tag(filepath) given credentials (can be omitted)"""
        fs, path = _load_fs_and_path(tag, creds, session_creds=session_creds)
        fs: fsspec.AbstractFileSystem = fs

        if (
            fs.exists(path)
            and not fs.exists(f"{path}/meta.json")
            and not fs.exists(f"{path}/HUB_DATASET")
            and len(fs.ls(path, detail=False)) > 0
        ):
            raise Exception(f"This path {path} is not a dataset path, tag: {tag}")
        self.delete(tag, creds)
        fs.makedirs(path)

        with fs.open(f"{path}/HUB_DATASET", "w") as f:
            f.write("Hello World")

        tensor_paths = [f"{path}/{t}" for t in self._tensors]
        for tensor_path in tensor_paths:
            fs.makedirs(tensor_path)
        tensor_meta = {
            name: _preprocess_meta_before_save(t._meta)
            for name, t in self._tensors.items()
        }
        count = self.count
        try:
            if count == -1:
                count = self._store_unknown_sized_ds(fs, path)
            else:
                self._store_known_sized_ds(fs, path)
        except PermissionError as e:
            logger.error(e)
            raise PermissionException(tag)

        for _, el in tensor_meta.items():
            el["shape"] = (count,) + tuple(el["shape"][1:])
        ds_meta = {"tensors": tensor_meta, "len": count}
        ds_info = dict()
        for key, value in self._metainfo.items():
            ds_info[key] = value
        ds_meta["metainfo"] = ds_info
        with fs.open(f"{path}/meta.json", "w") as f:
            f.write(json.dumps(ds_meta, indent=2, sort_keys=True))

        return load(tag, creds)

    def to_pytorch(self, transform=None, max_text_len=30):
        """
        Transforms into pytorch dataset

        Parameters
        ----------
        transform: func
            any transform that takes input a dictionary of a sample and returns transformed dictionary
        max_text_len: integer
            the maximum length of text strings that would be stored. Strings longer than this would be snipped
        """
        try:
            import torch

            global torch
        except ImportError:
            pass
        return TorchDataset(self, transform, max_text_len)

    def to_tensorflow(self, max_text_len=30):
        """
        Transforms into tensorflow dataset

        Parameters
        ----------
        max_text_len: integer
            the maximum length of text strings that would be stored. Strings longer than this would be snipped

        """
        try:
            import tensorflow as tf
        except ImportError:
            pass

        def tf_gen(step=4):
            with dask.config.set(scheduler="sync"):
                for index in range(0, len(self), step):
                    arrs = [self[index : index + step].values() for i in range(1)]
                    arrs = list(map(lambda x: x._array, _flatten(arrs)))
                    arrs = dask.delayed(list, pure=False, nout=len(list(self.keys())))(
                        arrs
                    )
                    arrs = arrs.compute()
                    for ind, arr in enumerate(arrs):
                        if arr.dtype.type is np.str_:
                            arr = [
                                ([ord(x) for x in sample.tolist()[0:max_text_len]])
                                for sample in arr
                            ]
                            arr = np.array(
                                [
                                    np.pad(
                                        sample,
                                        (0, max_text_len - len(sample)),
                                        "constant",
                                        constant_values=(32),
                                    )
                                    for sample in arr
                                ]
                            )
                            arrs[ind] = arr

                    for i in range(step):
                        sample = {key: r[i] for key, r in zip(self[index].keys(), arrs)}
                        yield sample

        def tf_dtype(np_dtype):
            try:
                if "U" in np_dtype:
                    return tf.dtypes.as_dtype("string")
                return tf.dtypes.as_dtype(np_dtype)
            except Exception as e:
                logger.log(e)
                return tf.variant

        output_shapes = {}
        output_types = {}
        for key in self.keys():
            output_types[key] = tf_dtype(self._tensors[key].dtype)
            output_shapes[key] = self._tensors[key].shape[1:]

            # if this is a string, we change the type to int, as it's going to become ascii. shape is also set to None
            if output_types[key] == tf.dtypes.as_dtype("string"):
                output_types[key] = tf.dtypes.as_dtype("int8")
                output_shapes[key] = None

        # TODO use None for dimensions you don't know the length tf.TensorShape([None])
        # FIXME Dataset Generator is not very good with multiprocessing but its good for fast tensorflow support
        return tf.data.Dataset.from_generator(
            tf_gen,
            output_types=output_types,
            # output_shapes=output_shapes,
        )


def _numpy_load(
    fs: fsspec.AbstractFileSystem, filepath: str, codec: BaseCodec
) -> np.ndarray:
    """Given filesystem and filepath, loads numpy array"""
    # assert fs.exists(
    #    filepath
    # ), f"Dataset file {filepath} does not exists. Your dataset data is likely to be corrupted"

    try:
        with fs.open(filepath, "rb") as f:
            return codec.decode(f.read())
    except Exception as e:
        logger.error(traceback.format_exc() + str(e))
        raise Exception(
            f"Dataset file {filepath} does not exists. Your dataset data is likely to be corrupted"
        )


def get_text(input):
    """Converts strings stored as ascii value tensors back into strings"""
    if input.ndim == 1:
        try:
            text = "".join([chr(x) for x in input]).rstrip()
            return text
        except Exception as e:
            logger.error(traceback.format_exc() + str(e))
            raise Exception(
                "get_text can only be called on a tensor of text or a batch of tensors of text"
            )
    elif input.ndim == 2:
        try:
            text = ["".join([chr(x) for x in sample]).rstrip() for sample in input]
            return text
        except Exception as e:
            logger.error(traceback.format_exc() + str(e))
            raise Exception(
                "get_text can only be called on a tensor of text or a batch of tensors of text"
            )
    else:
        raise Exception(
            f"Got input of dimension {input.ndim} for get_text. Expected dimension of 1 or 2"
        )


def load(tag, creds=None, session_creds=True) -> Dataset:
    """Load a dataset from repository using given url and credentials (optional)"""
    fs, path = _load_fs_and_path(tag, creds, session_creds=session_creds)
    fs: fsspec.AbstractFileSystem = fs
    path_2 = f"{path}/meta.json"
    if not fs.exists(path_2):
        raise HubDatasetNotFoundException(tag)

    with fs.open(path_2, "r") as f:
        ds_meta = json.loads(f.read())

    for name in ds_meta["tensors"]:
        assert fs.exists(
            f"{path}/{name}"
        ), f"Tensor {name} of {tag} dataset does not exist"

    if "dask" not in sys.modules:
        raise ModuleNotInstalledException("dask")
    else:
        import dask
        import dask.array

        global dask

    if ds_meta["len"] == 0:
        logger.warning("The dataset is empty (has 0 samples)")

        return Dataset(
            {
                name: Tensor(
                    tmeta,
                    dask.array.from_array(
                        np.empty(shape=(0,) + tuple(tmeta["shape"][1:]), dtype="uint8"),
                    ),
                )
                for name, tmeta in ds_meta["tensors"].items()
            },
            metainfo=ds_meta.get("metainfo"),
        )
    len_ = ds_meta["len"]

    # added reverse compatibility for previous versions
    for name, tmeta in ds_meta["tensors"].items():
        if "chunksize" not in tmeta:
            tmeta["chunksize"] = 1

    return Dataset(
        {
            name: Tensor(
                tmeta,
                _dask_concat(
                    [
                        dask.array.from_delayed(
                            dask.delayed(_numpy_load)(
                                fs,
                                f"{path}/{name}/{i}.npy",
                                codec_from_name(tmeta.get("dcompress")),
                            ),
                            shape=(min(tmeta["chunksize"], len_ - i),)
                            + tuple(tmeta["shape"][1:]),
                            dtype=tmeta["dtype"],
                        )
                        for i in range(0, len_, tmeta["chunksize"])
                    ]
                ),
            )
            for name, tmeta in ds_meta["tensors"].items()
        },
        metainfo=ds_meta.get("metainfo"),
    )


def _is_arraylike(arr):
    return (
        isinstance(arr, np.ndarray) or isinstance(arr, list) or isinstance(arr, tuple)
    )


def _is_tensor_dynamic(tensor):
    arr = tensor._array.to_delayed().flatten()[0].compute()
    return str(tensor.dtype) == "object" and _is_arraylike(arr.flatten()[0])


class TorchDataset:
    def __init__(self, ds, transform=None, max_text_len=30):
        self._ds = ds
        self._transform = transform
        self._dynkeys = {
            key for key in self._ds.keys() if _is_tensor_dynamic(self._ds[key])
        }
        self._max_text_len = max_text_len

        def cost(nbytes, time):
            print(nbytes, time)
            return float(time) / (nbytes or 1) / 1e9

        self.client = None

    def _do_transform(self, data):
        return self._transform(data) if self._transform else data

    def __len__(self):
        return len(self._ds)

    def __getitem__(self, index):
        with dask.config.set(scheduler="sync", delayed_pure=True):
            arrs = [self._ds[index : index + 1].values() for i in range(1)]
            arrs = list(map(lambda x: x._array, _flatten(arrs)))
            arrs = dask.delayed(list, pure=True, nout=len(list(self._ds.keys())))(arrs)
            arrs = arrs.compute()

            arrs = {key: r[0] for key, r in zip(self._ds[index].keys(), arrs)}

        objs = self._do_transform(arrs)
        if isinstance(objs, dict):
            objs = {k: self._to_tensor(k, v) for k, v in objs.items()}
        elif isinstance(objs, list):
            objs = [self._to_tensor(v) for v in objs]
        return objs

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def _to_tensor(self, key, sample):
        if key not in self._dynkeys:
            if isinstance(sample, np.str_):
                sample = np.array(
                    [ord(x) for x in sample.tolist()[0 : self._max_text_len]]
                )
                sample = np.pad(
                    sample,
                    (0, self._max_text_len - len(sample)),
                    "constant",
                    constant_values=(32),
                )
            return torch.tensor(sample)
        else:
            return [torch.tensor(item) for item in sample]

    def collate_fn(self, batch):
        batch = tuple(batch)
        keys = tuple(batch[0].keys())
        ans = {key: [item[key] for item in batch] for key in keys}

        for key in keys:
            if key not in self._dynkeys:
                ans[key] = torch.stack(ans[key], dim=0, out=None)

        return ans


def _dask_concat(arr):
    if len(arr) == 1:
        return arr[0]
    else:
        return dask.array.concatenate(arr)



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from hub.store.store import get_cache_path


def test_get_cache_path():
    cache_folder = "./cache/"
    assert "./cache/test/testdb" == get_cache_path("s3://test/testdb", cache_folder)
    assert "./cache/test/testdb" == get_cache_path("gcs://test/testdb", cache_folder)
    assert "./cache/test/testdb" == get_cache_path("test/testdb", cache_folder)
    assert "./cache/test/testdb" == get_cache_path("https://test/testdb", cache_folder)
    assert "./cache/test/testdb" == get_cache_path("~/test/testdb", cache_folder)
    assert "./cache/test/testdb" == get_cache_path("/test/testdb", cache_folder)
    assert "./cache/test\\testdb" == get_cache_path("C:\\test\\testdb", cache_folder)


if __name__ == "__main__":
    test_get_cache_path()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from .features import Tensor, Primitive, SchemaDict
from .audio import Audio
from .bbox import BBox
from .class_label import ClassLabel
from .image import Image
from .mask import Mask
from .polygon import Polygon
from .segmentation import Segmentation
from .sequence import Sequence
from .text import Text
from .video import Video



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import zarr
import numpy as np
import math
from typing import Dict, Iterable
from hub.api.dataset import Dataset
from tqdm import tqdm
from collections.abc import MutableMapping
from hub.utils import batchify
from hub.api.dataset_utils import get_value, slice_split, str_to_int, slice_extract_info
import collections.abc as abc
from hub.api.datasetview import DatasetView
from pathos.pools import ProcessPool, ThreadPool
from hub.schema.sequence import Sequence
from hub.schema.features import featurify
import os
from hub.defaults import OBJECT_CHUNK


def get_sample_size(schema, workers):
    """Given Schema, decides how many samples to take at once and returns it"""
    schema = featurify(schema)
    samples = 10000
    for feature in schema._flatten():
        shp = list(feature.max_shape)
        if len(shp) == 0:
            shp = [1]

        sz = np.dtype(feature.dtype).itemsize
        if feature.dtype == "object":
            sz = (16 * 1024 * 1024 * 8) / 128

        def prod(shp):
            res = 1
            for s in shp:
                res *= s
            return res

        samples = min(samples, (16 * 1024 * 1024 * 8) // (prod(shp) * sz))
    samples = max(samples, 1)
    return samples * workers


class Transform:
    def __init__(
        self, func, schema, ds, scheduler: str = "single", workers: int = 1, **kwargs
    ):
        """| Transform applies a user defined function to each sample in single threaded manner.

        Parameters
        ----------
        func: function
            user defined function func(x, **kwargs)
        schema: dict of dtypes
            the structure of the final dataset that will be created
        ds: Iterative
            input dataset or a list that can be iterated
        scheduler: str
            choice between "single", "threaded", "processed"
        workers: int
            how many threads or processes to use
        **kwargs:
            additional arguments that will be passed to func as static argument for all samples
        """
        self._func = func
        self.schema = schema
        self._ds = ds
        self.kwargs = kwargs
        self.workers = workers

        if isinstance(self._ds, Transform):
            self.base_ds = self._ds.base_ds
            self._func = self._ds._func[:]
            self._func.append(func)
            self.kwargs = self._ds.kwargs[:]
            self.kwargs.append(kwargs)
        else:
            self.base_ds = ds
            self._func = [func]
            self.kwargs = [kwargs]

        if scheduler == "threaded" or (scheduler == "single" and workers > 1):
            self.map = ThreadPool(nodes=workers).map
        elif scheduler == "processed":
            self.map = ProcessPool(nodes=workers).map
        elif scheduler == "single":
            self.map = map
        elif scheduler == "ray":
            try:
                from ray.util.multiprocessing import Pool as RayPool
            except Exception:
                pass
            self.map = RayPool().map
        else:
            raise Exception(
                f"Scheduler {scheduler} not understood, please use 'single', 'threaded', 'processed'"
            )

    def __len__(self):
        return self.shape[0]

    def __getitem__(self, slice_):
        """| Get an item to be computed without iterating on the whole dataset.
        | Creates a dataset view, then a temporary dataset to apply the transform.
        Parameters:
        ----------
        slice_: slice
            Gets a slice or slices from dataset
        """
        if not isinstance(slice_, abc.Iterable) or isinstance(slice_, str):
            slice_ = [slice_]

        slice_ = list(slice_)
        subpath, slice_list = slice_split(slice_)

        slice_list = slice_list or [slice(None, None, None)]

        num, ofs = slice_extract_info(slice_list[0], self.shape[0])
        ds_view = self._ds[slice_list[0]]

        path = os.path.expanduser("~/.activeloop/tmparray")
        new_ds = self.store(path, length=num, ds=ds_view, progressbar=False)

        index = 1 if len(slice_) > 1 else 0
        slice_[index] = (
            slice(None, None, None) if not isinstance(slice_list[0], int) else 0
        )  # Get all shape dimension since we already sliced
        return new_ds[slice_]

    def __iter__(self):
        for index in range(len(self)):
            yield self[index]

    @classmethod
    def _flatten_dict(self, d: Dict, parent_key="", schema=None):
        """| Helper function to flatten dictionary of a recursive tensor

        Parameters
        ----------
        d: dict
        """
        items = []
        for k, v in d.items():
            new_key = parent_key + "/" + k if parent_key else k
            if isinstance(v, MutableMapping) and not isinstance(
                self.dtype_from_path(new_key, schema), Sequence
            ):
                items.extend(
                    self._flatten_dict(v, parent_key=new_key, schema=schema).items()
                )
            else:
                items.append((new_key, v))
        return dict(items)

    @classmethod
    def dtype_from_path(cls, path, schema):
        """
        Helper function to get the dtype from the path
        """
        path = path.split("/")
        cur_type = schema
        for subpath in path[:-1]:
            cur_type = cur_type[subpath]
            cur_type = cur_type.dict_
        return cur_type[path[-1]]

    @classmethod
    def _unwrap(cls, results):
        """
        If there is any list then unwrap it into its elements
        """
        items = []
        for r in results:
            if isinstance(r, dict):
                items.append(r)
            else:
                items.extend(r)
        return items

    def _split_list_to_dicts(self, xs):
        """| Helper function that transform list of dicts into dicts of lists

        Parameters
        ----------
        xs: list of dicts
        Returns
        ----------
        xs_new: dicts of lists
        """
        xs_new = {}
        for x in xs:
            if isinstance(x, list):
                x = dict(
                    zip(self._flatten_dict(self.schema, schema=self.schema).keys(), x)
                )

            for key, value in x.items():
                if key in xs_new:
                    xs_new[key].append(value)
                else:
                    xs_new[key] = [value]
        return xs_new

    def _pbar(self, show: bool = True):
        """
        Returns a progress bar, if empty then it function does nothing
        """

        def _empty_pbar(xs, **kwargs):
            return xs

        single_threaded = self.map == map
        return tqdm if show and single_threaded else _empty_pbar

    def create_dataset(
        self, url: str, length: int = None, token: dict = None, public: bool = True
    ):
        """Helper function to create a dataset"""
        shape = (length,)
        ds = Dataset(
            url,
            mode="w",
            shape=shape,
            schema=self.schema,
            token=token,
            fs=zarr.storage.MemoryStore() if "tmp" in url else None,
            cache=False,
            public=public,
        )
        return ds

    def upload(self, results, ds: Dataset, token: dict, progressbar: bool = True):
        """Batchified upload of results.
        For each tensor batchify based on its chunk and upload.
        If tensor is dynamic then still upload element by element.
        For dynamic tensors, it disable dynamicness and then enables it back.

        Parameters
        ----------
        dataset: hub.Dataset
            Dataset object that should be written to
        results:
            Output of transform function
        progressbar: bool
        Returns
        ----------
        ds: hub.Dataset
            Uploaded dataset
        """
        offset = ds.indexes[
            0
        ]  # here ds.indexes will always be a contiguous list as obtained after slicing
        for key, value in results.items():
            chunk = ds[key].chunksize[0]
            chunk = 1 if chunk == 0 else chunk
            value = get_value(value)
            value = str_to_int(value, ds.dataset.tokenizer)

            num_chunks = len(value) / (chunk * self.workers)
            num_chunks = (
                1 + int(num_chunks) if num_chunks != int(num_chunks) else num_chunks
            )
            length = int(num_chunks * chunk) if self.workers != 1 else len(value)
            batched_values = (
                batchify(value, length, length + ((chunk - (offset % chunk))) % chunk)
                if length != len(value)
                else batchify(value, length)
            )
            len_batches = [len(item) for item in batched_values]

            def upload_chunk(i_batch):
                i, batch = i_batch
                length = len(batch)
                cur_offset = 0
                for it in range(i):
                    cur_offset += len_batches[it]
                slice_ = slice(cur_offset, cur_offset + length)
                ds[key, slice_] = batch

            index_batched_values = list(
                zip(list(range(len(batched_values))), batched_values)
            )

            # Disable dynamic arrays
            ds.dataset._tensors[f"/{key}"].disable_dynamicness()
            list(map(upload_chunk, index_batched_values))

            # Enable and rewrite shapes
            if ds.dataset._tensors[f"/{key}"].is_dynamic:
                ds.dataset._tensors[f"/{key}"].enable_dynamicness()
                ds.dataset._tensors[f"/{key}"].set_shape(
                    [slice(offset, offset + len(value))], value
                )

        ds.flush()
        return ds

    def call_func(self, fn_index, item, as_list=False):
        """Calls all the functions one after the other

        Parameters
        ----------
        fn_index: int
            The index starting from which the functions need to be called
        item:
            The item on which functions need to be applied
        as_list: bool, optional
            If true then treats the item as a list.

        Returns
        ----------
        result:
            The final output obtained after all transforms
        """
        result = item
        if fn_index < len(self._func):
            if as_list:
                result = [self.call_func(fn_index, it) for it in result]
            else:
                result = self._func[fn_index](result, **self.kwargs[fn_index])
                result = self.call_func(fn_index + 1, result, isinstance(result, list))
        result = self._unwrap(result) if isinstance(result, list) else result
        return result

    def store_shard(self, ds_in: Iterable, ds_out: Dataset, offset: int, token=None):
        """
        Takes a shard of iteratable ds_in, compute and stores in DatasetView
        """

        def _func_argd(item):
            if isinstance(item, DatasetView) or isinstance(item, Dataset):
                item = item.numpy()
            result = self.call_func(
                0, item
            )  # If the iterable obtained from iterating ds_in is a list, it is not treated as list
            return result

        ds_in = list(ds_in)
        results = self.map(
            _func_argd,
            ds_in,
        )
        results = self._unwrap(results)
        results = self.map(lambda x: self._flatten_dict(x, schema=self.schema), results)
        results = list(results)

        results = self._split_list_to_dicts(results)

        results_values = list(results.values())
        if len(results_values) == 0:
            return 0

        n_results = len(results_values[0])
        if n_results == 0:
            return 0

        additional = max(offset + n_results - ds_out.shape[0], 0)
        ds_out.append_shape(additional)

        self.upload(
            results,
            ds_out[offset : offset + n_results],
            token=token,
        )

        return n_results

    def store(
        self,
        url: str,
        token: dict = None,
        length: int = None,
        ds: Iterable = None,
        progressbar: bool = True,
        sample_per_shard: int = None,
        public: bool = True,
    ):
        """| The function to apply the transformation for each element in batchified manner

        Parameters
        ----------
        url: str
            path where the data is going to be stored
        token: str or dict, optional
            If url is referring to a place where authorization is required,
            token is the parameter to pass the credentials, it can be filepath or dict
        length: int
            in case shape is None, user can provide length
        ds: Iterable
        progressbar: bool
            Show progress bar
        sample_per_shard: int
            How to split the iterator not to overfill RAM
        public: bool, optional
            only applicable if using hub storage, ignored otherwise
            setting this to False allows only the user who created it to access the dataset and
            the dataset won't be visible in the visualizer to the public
        Returns
        ----------
        ds: hub.Dataset
            uploaded dataset
        """

        ds_in = ds or self.base_ds

        # compute shard length
        if sample_per_shard is None:
            n_samples = get_sample_size(self.schema, self.workers)
        else:
            n_samples = sample_per_shard
        try:
            length = len(ds_in) if hasattr(ds_in, "__len__") else n_samples
        except Exception:
            length = length or n_samples

        if length < n_samples:
            n_samples = length

        ds_out = self.create_dataset(url, length=length, token=token, public=public)

        def batchify_generator(iterator: Iterable, size: int):
            batch = []
            for el in iterator:
                batch.append(el)
                if len(batch) >= size:
                    yield batch
                    batch = []
            yield batch

        start = 0
        total = 0

        with tqdm(
            total=length,
            unit_scale=True,
            unit=" items",
            desc=f"Computing the transformation in chunks of size {n_samples}",
        ) as pbar:
            for ds_in_shard in batchify_generator(ds_in, n_samples):
                n_results = self.store_shard(ds_in_shard, ds_out, start, token=token)
                total += n_results
                pbar.update(len(ds_in_shard))
                start += n_results

        ds_out.resize_shape(total)
        ds_out.flush()
        return ds_out

    @property
    def shape(self):
        return self._ds.shape



import os

import hub
import numpy as np
from hub.auto import util
from hub.auto.infer import state
from PIL import Image
from tqdm import tqdm

USE_TQDM = True


@state.directory_parser(priority=0)
def image_classification(path, scheduler, workers):
    children = util.get_children(path, only_dirs=True)

    # check if there is >= 2 children (means there are at least 2 folders)
    if len(children) < 2:
        print("test1")
        return None

    # check if children's contents has image files
    for child in children:
        if not util.files_are_of_extension(child, util.IMAGE_EXTS):
            print("test2")
            return None

    # parse dataset
    data = []
    class_names = set()
    max_shape = np.zeros(3, dtype=np.uint8)  # CHW
    all_same_shape = True
    image_shape = None  # if shape is all the same, use that instead of max_shape
    for child in tqdm(
        children,
        desc="parsing image classification dataset",
        total=len(children),
        disable=not USE_TQDM,
    ):
        label = os.path.basename(child)

        filepaths = util.get_children(child)
        for filepath in filepaths:
            # ignore non-image extension files
            if util.get_ext(filepath) not in util.IMAGE_EXTS:
                continue

            shape = np.array(util.get_image_shape(filepath))

            # check if all images have the same shape
            if all_same_shape:
                if image_shape is None:
                    image_shape = shape
                elif not np.array_equal(image_shape, shape):
                    all_same_shape = False

            max_shape = np.maximum(max_shape, shape)
            data.append((filepath, label.lower()))
            class_names.add(label.lower())

    # create schema
    class_names = list(sorted(list(class_names)))
    max_shape = tuple([int(x) for x in max_shape])
    actual_shape = (
        tuple([int(x) for x in image_shape]) if all_same_shape else (None, None, None)
    )
    max_shape = None if all_same_shape else max_shape
    schema = {
        "image": hub.schema.Image(
            shape=actual_shape, dtype="uint8", max_shape=max_shape
        ),
        "label": hub.schema.ClassLabel(names=class_names),
    }

    # create transform for putting data into hub format
    @hub.transform(schema=schema, scheduler=scheduler, workers=workers)
    def upload_data(sample):
        path = sample[0]
        img = Image.open(path)
        img = np.asarray(img)
        label = class_names.index(sample[1])
        return {"image": img, "label": label}

    return upload_data(data)



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import json
import os
import time
from hub import config
from hub.client.base import HubHttpClient
from hub.client.auth import AuthClient
from pathlib import Path

from hub.exceptions import NotFoundException
from hub.log import logger
import traceback


class HubControlClient(HubHttpClient):
    """
    Controlling Hub through rest api
    """

    def __init__(self):
        super().__init__()
        self.details = self.get_config()

    def get_dataset_path(self, tag):
        try:
            dataset = self.request(
                "GET",
                config.GET_DATASET_PATH_SUFFIX,
                params={"tag": tag},
                endpoint=config.HUB_REST_ENDPOINT,
            ).json()

        except NotFoundException:
            dataset = None
        return dataset

    def get_credentials(self):

        if self.auth_header is None:
            token = AuthClient().get_access_token(username="public", password="")
            self.auth_header = f"Bearer {token}"

        r = self.request(
            "GET",
            config.GET_CREDENTIALS_SUFFIX,
            endpoint=config.HUB_REST_ENDPOINT,
        ).json()

        details = {
            "_id": r["_id"],
            "region": r["region"],
            "session_token": r["session_token"],
            "access_key": r["access_key"],
            "secret_key": r["secret_key"],
            "endpoint": r["endpoint"],
            "expiration": r["expiration"],
            "bucket": r["bucket"],
        }

        self.save_config(details)
        return details

    def get_config(self, reset=False):

        if not os.path.isfile(config.STORE_CONFIG_PATH) or self.auth_header is None:
            self.get_credentials()

        with open(config.STORE_CONFIG_PATH) as file:
            details = file.readlines()
            details = json.loads("".join(details))

        if float(details["expiration"]) < time.time() - 36000 or reset:
            details = self.get_credentials()
        return details

    def save_config(self, details):
        path = Path(config.STORE_CONFIG_PATH)
        os.makedirs(path.parent, exist_ok=True)
        with open(config.STORE_CONFIG_PATH, "w") as file:
            file.writelines(json.dumps(details))

    def create_dataset_entry(self, username, dataset_name, meta, public=True):
        try:
            tag = f"{username}/{dataset_name}"
            repo = f"public/{username}" if public else f"private/{username}"

            self.request(
                "POST",
                config.CREATE_DATASET_SUFFIX,
                json={
                    "tag": tag,
                    "repository": repo,
                    "public": public,
                    "rewrite": True,
                },
                endpoint=config.HUB_REST_ENDPOINT,
            ).json()
        except Exception as e:
            logger.error(
                "Unable to create Dataset entry" + traceback.format_exc() + str(e)
            )

    def update_dataset_state(self, username, dataset_name, state, progress=0):
        try:
            tag = f"{username}/{dataset_name}"
            self.request(
                "POST",
                config.UPDATE_STATE_SUFFIX,
                json={
                    "tag": tag,
                    "state": state,
                    "progress": progress,
                },
                endpoint=config.HUB_REST_ENDPOINT,
            ).json()
        except Exception as e:
            logger.error(
                "Unable to update Dataset entry state "
                + traceback.format_exc()
                + str(e)
            )

    def delete_dataset_entry(self, username, dataset_name):
        try:
            tag = f"{username}/{dataset_name}"
            suffix = f"{config.DATASET_SUFFIX}/{tag}"
            self.request(
                "DELETE",
                suffix,
                endpoint=config.HUB_REST_ENDPOINT,
            ).json()
        except Exception as e:
            logger.error(
                "Unable to delete Dataset entry" + traceback.format_exc() + str(e)
            )



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from hub.utils import _tuple_product
from hub.numcodecs import PngCodec
import math

import numpy as np

from hub.defaults import CHUNK_DEFAULT_SIZE, OBJECT_CHUNK, DEFAULT_COMPRESSOR
from hub.exceptions import HubException


class ShapeDetector:
    # TODO replace asserts with appropriate errors, not clear what to do if it misses the alert
    # TODO move this to hub/store
    def __init__(
        self,
        shape,
        max_shape=None,
        chunks=None,
        dtype="float64",
        chunksize=CHUNK_DEFAULT_SIZE,
        object_chunking=OBJECT_CHUNK,
        compressor=DEFAULT_COMPRESSOR,
    ):
        self._int32max = np.iinfo(np.dtype("int32")).max

        self._dtype = dtype = np.dtype(dtype)
        self._object_chunking = object_chunking
        self._compressor = compressor

        self._chunksize = chunksize = self._get_chunksize(chunksize, compressor)
        self._shape = shape = self._get_shape(shape)
        self._max_shape = max_shape = self._get_max_shape(shape, max_shape)
        self._chunks = chunks = self._get_chunks(
            shape, max_shape, chunks, dtype, chunksize
        )

    def _get_chunksize(self, chunksize, compressor):
        if isinstance(compressor, PngCodec):
            return int(math.ceil(0.25 * chunksize))
        else:
            return chunksize

    def _get_shape(self, shape):
        assert shape is not None
        shape = (shape,) if isinstance(shape, int) else tuple(shape)
        for s in shape:
            assert s is None or isinstance(s, int)
        assert isinstance(shape[0], int)
        return shape

    def _get_max_shape(self, shape, max_shape):
        if max_shape is None:
            return tuple([s or self._int32max for s in shape])
        elif isinstance(max_shape, int):
            assert max_shape == shape[0]
            return self._get_max_shape(shape, None)
        else:
            max_shape = tuple(max_shape)
            assert len(shape) == len(max_shape)
            for (s, ms) in zip(shape, max_shape):
                if not isinstance(ms, int):
                    raise HubException("MaxShape Dimension should be int")
                if s is not None and s != ms:
                    raise HubException(
                        """Dimension in shape cannot be != max_shape dimension, 
                        if shape is not None """
                    )
                assert s == ms or s is None and isinstance(ms, int)
            return max_shape

    def _get_chunks(self, shape, max_shape, chunks, dtype, chunksize):
        if chunks is None:
            prod = _tuple_product(max_shape[1:])
            if dtype == "object":
                return (self._object_chunking,) + max_shape[1:]
            if prod <= 2 * chunksize:
                # FIXME not properly handled object type, U type, and so on.
                sz = dtype.itemsize
                chunks = int(math.ceil(chunksize / (prod * sz)))
                return (chunks,) + max_shape[1:]
            else:
                return (1,) + self._determine_chunksizes(
                    max_shape[1:], dtype, chunksize
                )
        elif isinstance(chunks, int):
            assert chunks > 0
            if chunks > 1:
                return (chunks,) + tuple(
                    [s or ms for s, ms in zip(shape[1:], max_shape[1:])]
                )
            else:
                return (1,) + self._determine_chunksizes(
                    max_shape[1:], dtype, chunksize
                )
        else:
            chunks = tuple(chunks)
            if len(chunks) == 1:
                return self._get_chunks(shape, max_shape, chunks[0], dtype, chunksize)
            else:
                if len(chunks) != len(shape):
                    raise Exception(
                        "If you want a multidimensional chunk, the number of dimensions should match with that of the shape parameter."
                        f" Number of chunk dimensions received={len(chunks)}, number of chunk dimensions required={len(shape)}"
                    )
                if chunks[0] != 1:
                    raise Exception(
                        "If you want multiple samples in the chunk, then specify only the number in the chunk, rest of the dimensions "
                        f"should be omitted. In this case use 'chunks={chunks[0]}'"
                    )
                return chunks

    def _determine_chunksizes(self, max_shape, dtype, chunksize):
        """
        Autochunking of tensors
        Chunk is determined by 16MB blocks keeping left dimensions inside a chunk
        Dimensions from left are kept until 16MB block is filled

        Parameters
        ----------
        shape: tuple
            the shape of the whole array
        dtype: type
            the type of the element (int, float)
        chunk_default_size: int (optional)
            how big the chunk size should be. Default to 16MB (CHUNK_DEFAULT_SIZE)
        """

        sz = np.dtype(dtype).itemsize
        elem_count_in_chunk = chunksize / sz

        # Get left most part which will be left static inside the chunk
        a = list(max_shape)
        a.reverse()
        left_part = max_shape
        prod = 1
        for i, dim in enumerate(a):
            prod *= dim
            if elem_count_in_chunk < prod:
                left_part = max_shape[-i:]
                break

        # If the tensor is smaller then the chunk size return
        if len(left_part) == len(max_shape):
            return tuple(left_part)

        # Get the middle chunk size of dimension
        els = int(math.ceil(elem_count_in_chunk / np.prod(np.array(left_part))))

        # Contruct the chunksize shape
        chunksize = [els] + list(left_part)
        if len(chunksize) < len(max_shape):
            chunksize = [1] * (len(max_shape) - len(chunksize)) + chunksize
        return tuple(chunksize)

    @property
    def shape(self):
        return self._shape

    @property
    def max_shape(self):
        return self._max_shape

    @property
    def chunks(self):
        return self._chunks

    @property
    def dtype(self):
        return self._dtype

    @property
    def chunksize(self):
        return self._chunksize



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from hub.compute.transform import Transform
from hub.compute.ray import RayTransform, RayGeneratorTransform
from collections.abc import Iterable
from hub.exceptions import NotIterable


def transform(schema, scheduler="single", workers=1):
    """| Transform is a decorator of a function. The function should output a dictionary per sample.

    Parameters:
    ----------
    schema: Schema
        The output format of the transformed dataset
    scheduler: str
        "single" - for single threaded, "threaded" using multiple threads, "processed", "ray" scheduler, "dask" scheduler
    workers: int
        how many workers will be started for the process
    """

    def wrapper(func):
        def inner(ds, **kwargs):
            if not isinstance(ds, Iterable) and not isinstance(ds, str):
                raise NotIterable

            if scheduler == "ray":
                return RayTransform(
                    func, schema, ds, scheduler=scheduler, workers=workers, **kwargs
                )

            if scheduler == "ray_generator":
                return RayGeneratorTransform(
                    func, schema, ds, scheduler=scheduler, workers=workers, **kwargs
                )

            return Transform(
                func, schema, ds, scheduler=scheduler, workers=workers, **kwargs
            )

        return inner

    return wrapper



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import time
from collections import abc
from math import gcd

from numpy.lib.arraysetops import isin

from hub import defaults
from hub.exceptions import ShapeLengthException


def _flatten(list_):
    """
    Helper function to flatten the list
    """
    return [item for sublist in list_ for item in sublist]


def gcp_creds_exist():
    """Checks if credentials exists"""

    try:
        import os

        env = os.getenv("GOOGLE_APPLICATION_CREDENTIALS")
        if env is not None:
            return True
        from google.cloud import storage

        storage.Client()
    except Exception:
        return False
    return True


def s3_creds_exist():
    import boto3

    sts = boto3.client("sts")
    try:
        sts.get_caller_identity()
    except Exception:
        return False
    return True


def azure_creds_exist():
    """Checks if credentials exists"""

    import os

    env = os.getenv("ACCOUNT_KEY")
    return env is not None


def hub_creds_exist():
    """Checks if credentials exists"""

    import os

    env = os.getenv("ACTIVELOOP_HUB_PASSWORD")
    return env is not None


def minio_creds_exist():
    """Checks if credentials exists"""

    import os

    env1 = os.getenv("ACTIVELOOP_MINIO_KEY")
    env2 = os.getenv("ACTIVELOOP_MINIO_SECRET_ACCESS_KEY")
    return env1 is not None and env2 is not None


def pandas_loaded():
    try:
        import pandas as pd

        pd.__version__
    except ImportError:
        return False
    return True


def pytorch_loaded():
    try:
        import torch

        torch.__version__
    except ImportError:
        return False
    return True


def ray_loaded():
    try:
        import ray

        ray.__version__
    except ImportError:
        return False
    return True


def dask_loaded():
    try:
        import dask

        dask.__version__
    except ImportError:
        return False
    return True


def tensorflow_loaded():
    try:
        import tensorflow

        tensorflow.__version__
    except ImportError:
        return False
    return True


def tfds_loaded():
    try:
        import tensorflow_datasets

        tensorflow_datasets.__version__
    except ImportError:
        return False
    return True


def transformers_loaded():
    try:
        import transformers

        transformers.__version__
    except ImportError:
        return False
    return True


def pathos_loaded():
    try:
        import pathos

        pathos.__version__
    except ImportError:
        return False
    return True


def compute_lcm(a):
    """
    Lowest Common Multiple of a list a
    """
    if not a:
        return None
    lcm = a[0]
    for i in a[1:]:
        lcm = lcm * i // gcd(lcm, i)
    return int(lcm)


def batchify(iterable, n=1, initial=None):
    """
    Batchify an iteratable
    """
    ls = len(iterable)
    batches = []
    initial = initial or n
    batches.append(iterable[0 : min(initial, ls)])
    for ndx in range(initial, ls, n):
        batches.append(iterable[ndx : min(ndx + n, ls)])
    return batches


def _tuple_product(tuple_):
    res = 1
    for t in tuple_:
        res *= t
    return res


class Timer:
    def __init__(self, text):
        self._text = text

    def __enter__(self):
        self._start = time.time()

    def __exit__(self, *args):
        print(f"{self._text}: {time.time() - self._start}s")


def norm_shape(shape):
    shape = shape or (None,)
    if isinstance(shape, int):
        shape = (shape,)
    if not isinstance(shape, abc.Iterable):
        raise TypeError(
            f"shape is not None, int or Iterable, type(shape): {type(shape)}"
        )
    shape = tuple(shape)
    if not all([isinstance(s, int) or s is None for s in shape]):
        raise TypeError(f"Shape elements can be either int or None | shape: {shape}")
    return shape


def norm_cache(cache):
    cache = cache or 0
    if not isinstance(cache, int):
        raise TypeError("Cache should be None or int")
    return cache



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import numpy as np
import pytest

from .numcodecs import PngCodec


@pytest.mark.parametrize("from_config", [False, True])
def test_png_codec(from_config: bool) -> None:
    codec = PngCodec()
    if from_config:
        config = codec.get_config()
        codec = PngCodec.from_config(config)
    arr = np.ones((1000, 1000, 3), dtype="uint8")
    bytes_ = codec.encode(arr)
    arr_ = codec.decode(bytes_)
    assert (arr == arr_).all()


@pytest.mark.parametrize("solo_channel", [False, True])
def test_png_codec_config(solo_channel: bool) -> None:
    codec = PngCodec(solo_channel)
    config = codec.get_config()
    assert config["id"] == "png"
    assert config["solo_channel"] == solo_channel


@pytest.mark.parametrize("solo_channel", [False, True])
def test_png_codec_solo_channel(solo_channel: bool) -> None:
    codec = PngCodec(solo_channel)
    if solo_channel:
        arr = np.ones((1000, 1000, 1), dtype="uint8")
    else:
        arr = np.ones((1000, 1000), dtype="uint8")
    bytes_ = codec.encode(arr)
    arr_ = codec.decode(bytes_)
    assert (arr == arr_).all()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import numpy as np
import hub
import collections.abc as abc
from hub.api.dataset_utils import get_value, slice_split, str_to_int
from hub.exceptions import NoneValueException
import hub.api.objectview as objv


class TensorView:
    def __init__(
        self,
        dataset=None,
        subpath=None,
        slice_=None,
        lazy: bool = True,
    ):
        """Creates a TensorView object for a particular tensor in the dataset

        Parameters
        ----------
        dataset: hub.api.dataset.Dataset object
            The dataset whose TensorView is being created
        subpath: str
            The full path to the particular Tensor in the Dataset
        slice_: optional
            The `slice_` of this Tensor that needs to be accessed
        lazy: bool, optional
            Setting this to False will stop lazy computation and will allow items to be accessed without .compute()
        """

        if dataset is None:
            raise NoneValueException("dataset")
        if subpath is None:
            raise NoneValueException("subpath")

        self.dataset = dataset
        self.subpath = subpath
        self.lazy = lazy

        if isinstance(slice_, (int, slice)):
            self.slice_ = [slice_]
        elif isinstance(slice_, (tuple, list)):
            self.slice_ = list(slice_)
        self.nums = [None]
        self.offsets = [None]
        self.squeeze_dims = [False]
        self.indexes = self.slice_[0]  # int, slice or list
        self.is_contiguous = False
        if isinstance(self.indexes, list) and self.indexes:
            self.is_contiguous = self.indexes[-1] - self.indexes[0] + 1 == len(
                self.indexes
            )
        for it in self.slice_[1:]:
            if isinstance(it, int):
                self.nums.append(1)
                self.offsets.append(it)
                self.squeeze_dims.append(True)
            elif isinstance(it, slice):
                ofs = it.start or 0
                num = it.stop - ofs if it.stop else None
                self.nums.append(num)
                self.offsets.append(ofs)
                self.squeeze_dims.append(False)
        self.dtype = self.dtype_from_path(subpath)
        self.shape = self.dataset._tensors[self.subpath].get_shape(self.slice_)

    def numpy(self, label_name=False):
        """Gets the value from tensorview

        Parameters
        ----------
        label_name: bool, optional
            If the TensorView object is of the ClassLabel type, setting this to True would retrieve the label names
            instead of the label encoded integers, otherwise this parameter is ignored.
        """
        if isinstance(self.indexes, list):
            value = np.array(
                [
                    self.dataset._tensors[self.subpath][[index] + self.slice_[1:]]
                    for index in self.indexes
                ]
            )
        else:
            value = self.dataset._tensors[self.subpath][self.slice_]

        if isinstance(self.dtype, hub.schema.class_label.ClassLabel) and label_name:
            if isinstance(self.indexes, int):
                if value.ndim == 0:
                    value = self.dtype.int2str(value)
                elif value.ndim == 1:
                    value = [self.dtype.int2str(value[i]) for i in range(value.size)]
            else:
                if value.ndim == 1:
                    value = [self.dtype.int2str(value[i]) for i in range(value.size)]
                elif value.ndim == 2:
                    value = [
                        [self.dtype.int2str(item[i]) for i in range(item.size)]
                        for item in value
                    ]

        if isinstance(self.dtype, hub.schema.text.Text):
            if self.dataset.tokenizer is not None:
                from transformers import AutoTokenizer

                tokenizer = AutoTokenizer.from_pretrained("bert-base-cased")
                if value.ndim == 1:
                    return tokenizer.decode(value.tolist())
                elif value.ndim == 2:
                    return [tokenizer.decode(val.tolist()) for val in value]
            elif value.ndim == 1:
                return "".join(chr(it) for it in value.tolist())
            elif value.ndim == 2:
                return ["".join(chr(it) for it in val.tolist()) for val in value]
            raise ValueError(f"Unexpected value with shape for text {value.shape}")
        return value

    def compute(self, label_name=False):
        """Gets the value from tensorview

        Parameters
        ----------
        label_name: bool, optional
            If the TensorView object is of the ClassLabel type, setting this to True would retrieve the label names
            instead of the label encoded integers, otherwise this parameter is ignored.
        """
        return self.numpy(label_name=label_name)

    def __getitem__(self, slice_):
        """| Gets a slice or slices from tensorview
        | Usage:

        >>> images_tensorview = ds["image"]
        >>> return images_tensorview[7, 0:1920, 0:1080, 0:3].compute() # returns numpy array of 7th image
        """
        if not isinstance(slice_, abc.Iterable) or isinstance(slice_, str):
            slice_ = [slice_]
        slice_ = list(slice_)
        slice_ = self.slice_fill(slice_)
        subpath, slice_list = slice_split(slice_)
        new_nums = self.nums.copy()
        new_offsets = self.offsets.copy()
        if isinstance(self.indexes, list):
            new_indexes = self.indexes[slice_list[0]]
            if self.is_contiguous and new_indexes:
                new_indexes = slice(new_indexes[0], new_indexes[-1] + 1)
        elif isinstance(self.indexes, int):
            new_indexes = self.indexes
        else:
            ofs = self.indexes.start or 0
            num = self.indexes.stop - ofs if self.indexes.stop else None
            new_indexes = self._combine(slice_list[0], num, ofs)
        slice_list[0] = new_indexes
        # new_indexes = self.indexes[]
        if len(new_nums) < len(slice_list):
            new_nums.extend([None] * (len(slice_list) - len(new_nums)))
            new_offsets.extend([0] * (len(slice_list) - len(new_offsets)))
        for i in range(1, len(slice_list)):
            slice_list[i] = self._combine(slice_list[i], new_nums[i], new_offsets[i])
        for i in range(len(slice_list), len(new_nums)):
            cur_slice = (
                slice(new_offsets[i], new_offsets[i] + new_nums[i])
                if not self.squeeze_dims[i]
                else new_offsets[i]
            )
            slice_list.append(cur_slice)
        if subpath or (
            len(slice_list) > len(self.nums) and isinstance(self.dtype, objv.Sequence)
        ):
            objectview = objv.ObjectView(
                dataset=self.dataset,
                subpath=self.subpath + subpath,
                slice_=slice_list,
                lazy=self.lazy,
            )
            return objectview if self.lazy else objectview.compute()
        else:
            tensorview = TensorView(
                dataset=self.dataset,
                subpath=self.subpath,
                slice_=slice_list,
                lazy=self.lazy,
            )
            return tensorview if self.lazy else tensorview.compute()

    def __setitem__(self, slice_, value):
        """| Sets a slice or slices with a value
        | Usage:

        >>> images_tensorview = ds["image"]
        >>> images_tensorview[7, 0:1920, 0:1080, 0:3] = np.zeros((1920, 1080, 3), "uint8") # sets 7th image
        """
        self.dataset._auto_checkout()
        assign_value = get_value(value)
        # handling strings and bytes
        assign_value = str_to_int(assign_value, self.dataset.tokenizer)

        if not isinstance(slice_, abc.Iterable) or isinstance(slice_, str):
            slice_ = [slice_]
        slice_ = list(slice_)
        slice_ = self.slice_fill(slice_)
        subpath, slice_list = slice_split(slice_)
        if subpath:
            raise ValueError("Can't setitem of TensorView with subpath")
        new_nums = self.nums.copy()
        new_offsets = self.offsets.copy()
        if isinstance(self.indexes, list):
            new_indexes = self.indexes[slice_list[0]]
            if self.is_contiguous and new_indexes:
                new_indexes = slice(new_indexes[0], new_indexes[-1] + 1)
        elif isinstance(self.indexes, int):
            new_indexes = self.indexes
        else:
            ofs = self.indexes.start or 0
            num = self.indexes.stop - ofs if self.indexes.stop else None
            new_indexes = self._combine(slice_list[0], num, ofs)
        slice_list[0] = new_indexes
        if len(new_nums) < len(slice_list):
            new_nums.extend([None] * (len(slice_list) - len(new_nums)))
            new_offsets.extend([0] * (len(slice_list) - len(new_offsets)))
        for i in range(1, len(slice_list)):
            slice_list[i] = self._combine(slice_list[i], new_nums[i], new_offsets[i])
        for i in range(len(slice_list), len(new_nums)):
            cur_slice = (
                slice(new_offsets[i], new_offsets[i] + new_nums[i])
                if not self.squeeze_dims[i]
                else new_offsets[i]
            )
            slice_list.append(cur_slice)

        if isinstance(slice_list[0], (int, slice)):
            self.dataset._tensors[self.subpath][slice_list] = assign_value
        else:
            for i, index in enumerate(slice_list[0]):
                current_slice = [index] + slice_list[1:]
                self.dataset._tensors[subpath][current_slice] = assign_value[i]

    def _combine(self, slice_, num=None, ofs=0):
        "Combines a `slice_` with the current num and offset present in tensorview"
        if isinstance(slice_, int):
            self.check_slice_bounds(num=num, start=slice_)
            return ofs + slice_
        elif isinstance(slice_, slice):
            self.check_slice_bounds(
                num=num, start=slice_.start, stop=slice_.stop, step=slice_.step
            )
            if slice_.start is None and slice_.stop is None:
                return slice(ofs, None) if num is None else slice(ofs, ofs + num)
            elif slice_.stop is None:
                return (
                    slice(ofs + slice_.start, None)
                    if num is None
                    else slice(ofs + slice_.start, ofs + num)
                )
            elif slice_.start is None:
                return slice(ofs, ofs + slice_.stop)
            else:
                return slice(ofs + slice_.start, ofs + slice_.stop)
        else:
            raise TypeError(
                "type {} isn't supported in dataset slicing".format(type(slice_))
            )

    def check_slice_bounds(self, num=None, start=None, stop=None, step=None):
        "Checks whether the bounds of slice are in limits"
        if step and step < 0:  # negative step not supported
            raise ValueError("Negative step not supported in dataset slicing")
        if num and ((start and start >= num) or (stop and stop > num)):
            raise IndexError(
                "index out of bounds for dimension with length {}".format(num)
            )
        if start and stop and start > stop:
            raise IndexError("start index is greater than stop index")

    def dtype_from_path(self, path):
        "Gets the dtype of the Tensorview by traversing the schema"
        path = path.split("/")
        cur_type = self.dataset.schema.dict_
        for subpath in path[1:-1]:
            cur_type = cur_type[subpath]
            cur_type = cur_type.dict_
        return cur_type[path[-1]]

    def slice_fill(self, slice_):
        "Fills the slice with zeroes for the dimensions that have single elements and squeeze_dims true"
        if isinstance(self.indexes, int):
            new_slice_ = [0]
            offset = 0
        else:
            new_slice_ = [slice_[0]]
            offset = 1
        for i in range(1, len(self.nums)):
            if self.squeeze_dims[i]:
                new_slice_.append(0)
            elif offset < len(slice_):
                new_slice_.append(slice_[offset])
                offset += 1
        new_slice_ += slice_[offset:]
        return new_slice_

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return (
            "TensorView("
            + str(self.dtype)
            + ", subpath="
            + "'"
            + self.subpath
            + "', slice="
            + str(self.slice_)
            + ")"
        )

    def __iter__(self):
        """ Returns Iterable over samples """
        if isinstance(self.indexes, int):
            yield self
            return

        for i in range(len(self.indexes)):
            yield self[i]

    @property
    def chunksize(self):
        return self.dataset._tensors[self.subpath].chunksize

    @property
    def is_dynamic(self):
        return self.dataset._tensors[self.subpath].is_dynamic

    def disable_lazy(self):
        self.lazy = False

    def enable_lazy(self):
        self.lazy = True



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from typing import Tuple
from hub.schema.features import Tensor


class BBox(Tensor):
    """| HubSchema` for a normalized bounding box.

    Output: Tensor of type `float32` and shape `[4,]` which contains the
    normalized coordinates of the bounding box `[ymin, xmin, ymax, xmax]`


    Example: This example uploads a dataset with a Bounding box schema and retrieves it.

    ----------
    >>> import hub
    >>> from hub import Dataset, schema
    >>> from hub.schema import BBox
    >>> from numpy import asarray

    >>> tag = "username/dataset"
    >>>
    >>> # Create dataset
    >>> ds = Dataset(
    >>>   tag,
    >>>   shape=(10,),
    >>>   schema={
    >>>      "bbox": schema.BBox(dtype="uint8"),
    >>>  },
    >>> )
    >>>
    >>> ds["bbox", 1] = np.array([1,2,3,4])
    >>> ds.flush()

    >>> # Load data
    >>> ds = Dataset(tag)
    >>>
    >>> print(ds["bbox"][1].compute())
    [1 2 3 4]

    """

    def __init__(
        self,
        shape: Tuple[int, ...] = (4,),
        max_shape: Tuple[int, ...] = None,
        dtype="float64",
        chunks=None,
        compressor="lz4",
    ):
        """Construct the connector.

        Parameters
        ----------
        shape: tuple of ints or None
            The shape of bounding box.
            Will be (4,) if only one bounding box corresponding to each sample.
            If N bboxes corresponding to each sample, shape should be (N,)
            If the number of bboxes for each sample vary from 0 to M. The shape should be set to (None, 4) and max_shape should be set to (M, 4)
            Defaults to (4,).
        max_shape : Tuple[int], optional
            Maximum shape of BBox
        dtype : str
                dtype of bbox coordinates. Default: 'float32'
        chunks : Tuple[int] | True
            Describes how to split tensor dimensions into chunks (files) to store them efficiently.
            It is anticipated that each file should be ~16MB.
            Sample Count is also in the list of tensor's dimensions (first dimension)
            If default value is chosen, automatically detects how to split into chunks
        """
        self.check_shape(shape)
        super(BBox, self).__init__(
            shape=shape,
            max_shape=max_shape,
            dtype=dtype,
            chunks=chunks,
            compressor=compressor,
        )

    def __str__(self):
        out = super().__str__()
        out = "BBox" + out[6:]
        return out

    def __repr__(self):
        return self.__str__()

    def check_shape(self, shape):
        if len(shape) not in [1, 2] or shape[-1] != 4:
            raise ValueError(
                "Wrong BBox shape provided, should be of the format (4,) or (None, 4) or (N, 4)"
            )



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from hub.store.nested_store import NestedStore

import zarr


def test_nested_store():
    store = NestedStore(zarr.MemoryStore(), "hello")
    store["item"] = bytes("Hello World", "utf-8")
    assert store["item"] == bytes("Hello World", "utf-8")
    del store["item"]
    assert store.get("item") is None
    store["item1"] = bytes("Hello World 1", "utf-8")
    store["item2"] = bytes("Hello World 2", "utf-8")
    assert len(store) == 2
    assert tuple(store) == ("item1", "item2")
    try:
        store.close()
    except AttributeError as ex:
        assert "object has no attribute 'close'" in str(ex)


if __name__ == "__main__":
    test_nested_store()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from collections.abc import Iterable
from hub.api.dataset_utils import slice_split
from hub.api.compute_list import ComputeList


class ShardedDatasetView:
    def __init__(self, datasets: list) -> None:
        """
        | Creates a sharded simple dataset.
        | Datasets should have the schema.

        Parameters
        ----------
            datasets: list of Datasets
        """
        # TODO add schema check to make sure all datasets have the same schema

        self._schema = datasets[0].schema if datasets else None
        self.datasets = [
            ds
            if isinstance(ds.indexes, list)
            else ds.dataset[ds.indexes : ds.indexes + 1]
            for ds in datasets
        ]
        self.num_samples = sum([len(d) for d in self.datasets])

    @property
    def shape(self):
        return (self.num_samples,)

    def __len__(self):
        return self.num_samples

    def __str__(self):
        return f"ShardedDatasetView(shape={str(self.shape)})"

    def __repr__(self):
        return self.__str__()

    def identify_shard(self, index) -> tuple:
        """ Computes shard id and returns the shard index and offset """
        shard_id = 0
        count = 0
        for ds in self.datasets:
            count += len(ds)
            if index < count:
                return shard_id, count - len(ds)
            shard_id += 1
        return 0, 0

    def slicing(self, slice_list):
        """
        Identifies the dataset shard that should be used
        """
        shard_id, offset = self.identify_shard(slice_list[0])
        slice_list[0] = slice_list[0] - offset
        return slice_list, shard_id

    def __getitem__(self, slice_):
        if not isinstance(slice_, Iterable) or isinstance(slice_, str):
            slice_ = [slice_]
        slice_ = list(slice_)
        subpath, slice_list = slice_split(slice_)
        slice_list = slice_list or [slice(0, self.num_samples)]
        if isinstance(slice_list[0], int):
            # if integer it fetches the data from the corresponding dataset
            slice_list, shard_id = self.slicing(slice_list)
            slice_ = slice_list + [subpath] if subpath else slice_list
            return self.datasets[shard_id][slice_]
        else:
            # if slice it finds all the corresponding datasets included in the slice and generates tensorviews or datasetviews (depending on slice)
            # these views are stored in a ComputeList, calling compute on which will fetch data from all corresponding datasets and return a single result
            results = []
            cur_index = slice_list[0].start or 0
            cur_index = cur_index + self.num_samples if cur_index < 0 else cur_index
            cur_index = max(cur_index, 0)
            stop_index = slice_list[0].stop or self.num_samples
            stop_index = min(stop_index, self.num_samples)
            while cur_index < stop_index:
                shard_id, offset = self.identify_shard(cur_index)
                end_index = min(offset + len(self.datasets[shard_id]), stop_index)
                cur_slice_list = [
                    slice(cur_index - offset, end_index - offset)
                ] + slice_list[1:]
                current_slice = (
                    cur_slice_list + [subpath] if subpath else cur_slice_list
                )
                results.append(self.datasets[shard_id][current_slice])
                cur_index = end_index
            return ComputeList(results)

    def __setitem__(self, slice_, value) -> None:
        if not isinstance(slice_, Iterable) or isinstance(slice_, str):
            slice_ = [slice_]
        slice_ = list(slice_)
        subpath, slice_list = slice_split(slice_)
        slice_list = slice_list or [slice(0, self.num_samples)]
        if isinstance(slice_list[0], int):
            # if integer it assigns the data to the corresponding dataset
            slice_list, shard_id = self.slicing(slice_list)
            slice_ = slice_list + [subpath] if subpath else slice_list
            self.datasets[shard_id][slice_] = value
        else:
            # if slice it finds all the corresponding datasets and assigns slices of the value one by one
            cur_index = slice_list[0].start or 0
            cur_index = cur_index + self.num_samples if cur_index < 0 else cur_index
            cur_index = max(cur_index, 0)
            start_index = cur_index
            stop_index = slice_list[0].stop or self.num_samples
            stop_index = min(stop_index, self.num_samples)
            while cur_index < stop_index:
                shard_id, offset = self.identify_shard(cur_index)
                end_index = min(offset + len(self.datasets[shard_id]), stop_index)
                cur_slice_list = [
                    slice(cur_index - offset, end_index - offset)
                ] + slice_list[1:]
                current_slice = (
                    cur_slice_list + [subpath] if subpath else cur_slice_list
                )
                self.datasets[shard_id][current_slice] = value[
                    cur_index - start_index : end_index - start_index
                ]
                cur_index = end_index

    def __iter__(self):
        """ Returns Iterable over samples """
        for i in range(len(self)):
            yield self[i]

    @property
    def schema(self):
        return self._schema



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from typing import Tuple

import numpy as np

from hub.schema.features import Tensor


class Image(Tensor):
    """Schema for images would define the shape and structure for the dataset.

    Output: `Tensor` of type `uint8` and shape `[height, width, num_channels]`
    for BMP, JPEG, and PNG images

    Example: This example uploads an `image` to a Hub dataset `image_dataset` with `HubSchema` and retrieves it.

    ----------
    >>> import hub
    >>> from hub import Dataset, schema
    >>> from hub.schema import Image
    >>> from numpy import asarray

    >>> tag = "username/image_dataset"
    >>>
    >>> # Create dataset
    >>> ds=Dataset(
    >>>     tag,
    >>>     shape=(10,),
    >>>     schema={
    >>>         "image": schema.Image((height, width, 3), dtype="uint8"),
    >>>     },
    >>> )
    >>>
    >>> for index, image in enumerate(os.listdir("path/to/folder")):
    >>>         data = asarray(Image.open(image))
    >>>
    >>>         # Upload data
    >>>         ds["image"][index] = data
    >>>
    >>> ds.flush()

    >>> # Load data
    >>> ds = Dataset(tag)
    >>>
    >>> for i in range(len(ds)):
    >>>     print(ds["image"][i].compute())
    [[[124 112  64]
    [124 112  64]
    [124 112  64]
    ...
    [236 237 232]
    [238 239 234]
    [238 239 234]]]
    """

    def __init__(
        self,
        shape: Tuple[int, ...] = (None, None, 3),
        dtype="uint8",
        max_shape: Tuple[int, ...] = None,
        chunks=None,
        compressor="lz4",
    ):
        """| Construct the connector.

        Parameters
        ----------
        shape: tuple of ints or None
            The shape of decoded image:
            (height, width, channels) where height and width can be None.
            Defaults to (None, None, 3).
        dtype: `uint16` or `uint8` (default)
            `uint16` can be used only with png encoding_format
        encoding_format: 'jpeg' or 'png' (default)
             Format to serialize np.ndarray images on disk.
        max_shape : Tuple[int]
            Maximum shape of tensor shape if tensor is dynamic
        chunks : Tuple[int] | True
            Describes how to split tensor dimensions into chunks (files) to store them efficiently.
            It is anticipated that each file should be ~16MB.
            Sample Count is also in the list of tensor's dimensions (first dimension)
            If default value is chosen, automatically detects how to split into chunks


        Returns
        ----------
        `tf.Tensor` of type `tf.uint8` and shape `[height, width, num_channels]`
        for BMP, JPEG, and PNG images

        Raises
        ----------
        ValueError: If the shape, dtype or encoding formats are invalid

        """
        self._set_dtype(dtype)
        super().__init__(
            shape,
            dtype,
            max_shape=max_shape,
            chunks=chunks,
            compressor=compressor,
        )

    def _set_dtype(self, dtype):
        """Set the dtype."""
        dtype = str(np.dtype(dtype))
        if dtype not in ("uint8", "uint16"):
            raise ValueError(f"Not supported dtype for {self.__class__.__name__}")
        self.dtype = dtype

    def __str__(self):
        out = super().__str__()
        out = "Image" + out[6:]
        return out

    def __repr__(self):
        return self.__str__()



"""
License:
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import collections.abc as abc
from shutil import Error
from hub.schema.features import Shape
import json
import math

import numpy as np
from numpy.lib.arraysetops import isin
import zarr
import numcodecs

from hub.store.nested_store import NestedStore
from hub.store.shape_detector import ShapeDetector
from hub.defaults import DEFAULT_COMPRESSOR

from hub.exceptions import (
    DynamicTensorNotFoundException,
    ValueShapeError,
    DynamicTensorShapeException,
)
from hub.schema.sequence import Sequence


class DynamicTensor:
    """Class for handling dynamic tensor

    This class adds dynamic nature to storage tensor.
    The shape of tensor depends only on the index of the first dim.
    """

    # TODO Make first dim is extensible as well
    def __init__(
        self,
        fs_map: str,
        mode: str = "r",
        shape=None,
        max_shape=None,
        dtype="float64",
        chunks=None,
        compressor=DEFAULT_COMPRESSOR,
    ):
        """Constructor
        Parameters
        ----------
        fs_map : MutableMap
            Maps filesystem to MutableMap
        mode : str
            Mode in which tensor is opened (default is "r"), can be used to overwrite or append
        shape : Tuple[int | None]
            Shape of tensor, (must be specified) can contains Nones meaning the shape might change
        max_shape: Tuple[int | None]
            Maximum possible shape of the tensor (must be specified)
        dtype : str
            Numpy analog dtype for this tensor
        chunks : Tuple[int] | True
            How to split the tensor into chunks (files) (default is True)
            If chunks=True then chunksize will automatically be detected

        """
        if not (shape is None):
            # otherwise shape detector fails
            shapeDt = ShapeDetector(
                shape, max_shape, chunks, dtype, compressor=compressor
            )
            shape = shapeDt.shape
            max_shape = shapeDt.max_shape
            chunks = shapeDt.chunks
        elif "r" not in mode:
            raise TypeError("shape cannot be none")

        self.fs_map = fs_map
        exist_ = fs_map.get(".hub.dynamic_tensor")

        # if not exist_ and len(fs_map) > 0 and "w" in mode:
        #     raise OverwriteIsNotSafeException()
        exist = False if "w" in mode else exist_ is not None
        if "r" in mode and not exist:
            raise DynamicTensorNotFoundException()

        synchronizer = None
        # synchronizer = zarr.ThreadSynchronizer()
        # synchronizer = zarr.ProcessSynchronizer("~/activeloop/sync/example.sync")
        # if tensor exists and mode is read or append

        if ("r" in mode or "a" in mode) and exist:
            meta = json.loads(fs_map.get(".hub.dynamic_tensor").decode("utf-8"))
            shape = tuple(meta["shape"])
            self._dynamic_dims = get_dynamic_dims(shape)
            self._storage_tensor = zarr.open_array(
                store=fs_map, mode=mode, synchronizer=synchronizer
            )
            self._dynamic_tensor = (
                zarr.open_array(
                    NestedStore(fs_map, "--dynamic--"),
                    mode=mode,
                    synchronizer=synchronizer,
                )
                if self._dynamic_dims
                else None
            )
        # else we need to create or overwrite the tensor
        else:
            self._dynamic_dims = get_dynamic_dims(shape)
            self._storage_tensor = zarr.zeros(
                max_shape,
                dtype=dtype,
                chunks=chunks,
                store=fs_map,
                overwrite=("w" in mode),
                object_codec=numcodecs.Pickle(protocol=3)
                if str(dtype) == "object"
                else None,
                compressor=compressor,
                synchronizer=synchronizer,
            )
            self._dynamic_tensor = (
                zarr.zeros(
                    shape=(max_shape[0], len(self._dynamic_dims)),
                    mode=mode,
                    dtype=np.int32,
                    store=NestedStore(fs_map, "--dynamic--"),
                    synchronizer=synchronizer,
                    compressor=None,
                )
                if self._dynamic_dims
                else None
            )

            fs_map[".hub.dynamic_tensor"] = bytes(json.dumps({"shape": shape}), "utf-8")

        self.shape = shape
        self.max_shape = self._storage_tensor.shape
        self.chunks = self._storage_tensor.chunks
        self.dtype = self._storage_tensor.dtype

        if len(self.shape) != len(self.max_shape):
            raise DynamicTensorShapeException("length")
        for item in self.max_shape:
            if item is None:
                raise DynamicTensorShapeException("none")
        for item in zip(self.shape, self.max_shape):
            if item[0] is not None:
                if item[0] != item[1]:
                    raise DynamicTensorShapeException("not_equal")
        self._enabled_dynamicness = True

    def get_real_shape(self, slice_):
        slice_ = [slice_] if isinstance(slice_, int) else slice_
        if self._dynamic_tensor:
            if isinstance(slice_[0], int):
                real_shapes = self._dynamic_tensor[slice_[0]]
            else:
                start = slice_[0].start or 0
                end = slice_[0].stop or self.shape[0]
                real_shapes = [self.get_real_shape(i) for i in range(start, end)]
        else:
            real_shapes = None
        return real_shapes

    def __getitem__(self, slice_):
        """Gets a slice or slices from tensor"""
        if not isinstance(slice_, abc.Iterable):
            slice_ = [slice_]
        slice_ = list(slice_)
        # real_shapes is dynamic shapes based on first dim index, only dynamic dims are stored, static ones are ommitted
        real_shapes = self.get_real_shape(slice_)
        # Extend slice_ to dim count
        slice_ += [slice(0, None, 1) for i in self.max_shape[len(slice_) :]]
        if isinstance(real_shapes, list):
            start = slice_[0].start or 0
            slice_list = [
                self._get_slice([start + i] + slice_[1:], real_shapes[i])
                for i in range(len(real_shapes))
            ]
            return [self._storage_tensor[cur_slice] for cur_slice in slice_list]
        slice_ = self._get_slice(slice_, real_shapes)
        return self._storage_tensor[slice_]

    def __setitem__(self, slice_, value):
        """Sets a slice or slices with a value"""
        if not isinstance(slice_, abc.Iterable):
            slice_ = [slice_]
        slice_ = list(slice_)
        if self._dynamic_tensor and self._enabled_dynamicness:
            self.set_shape(slice_, value)
        slice_ += [slice(0, None, 1) for i in self.max_shape[len(slice_) :]]

        if self._dynamic_tensor and isinstance(slice_[0], int):
            real_shapes = self._dynamic_tensor[slice_[0]]
        elif self._dynamic_tensor and isinstance(slice_[0], slice):
            max_shape = value[0].shape
            for item in value:
                max_shape = tuple(max(value) for value in zip(max_shape, item.shape))
            for i in range(len(value)):
                pad = [
                    (0, max_shape[dim] - value[i].shape[dim])
                    for dim in range(value[i].ndim)
                ]
                value[i] = np.pad(value[i], pad)
            real_shapes = np.array(
                [
                    max_shape[i]
                    for i in range(len(max_shape))
                    if i + 1 in self._dynamic_dims
                ]
            )
        else:
            real_shapes = None

        if not self._enabled_dynamicness:
            real_shapes = (
                list(value.shape)
                if hasattr(value, "shape")
                else real_shapes
                if real_shapes is not None
                else [1]
            )

        slice_ = self._get_slice(slice_, real_shapes)
        value = self.check_value_shape(value, slice_)
        self._storage_tensor[slice_] = value

    def check_value_shape(self, value, slice_):
        """Checks if value can be set to the slice"""
        if None not in self.shape and self.dtype != "O":
            if not all([isinstance(sh, int) for sh in slice_]):
                expected_value_shape = tuple(
                    [
                        len(range(*slice_shape.indices(self.shape[i])))
                        for i, slice_shape in enumerate(slice_)
                        if not isinstance(slice_shape, int)
                    ]
                )

                if isinstance(value, list):
                    value = np.array(value)
                if isinstance(value, np.ndarray):
                    value_shape = [dim for dim in value.shape if dim != 1]
                    expected_shape = [dim for dim in expected_value_shape if dim != 1]
                    if value_shape != expected_shape:
                        raise ValueShapeError(expected_value_shape, value.shape)
                    else:
                        value = value.reshape(expected_value_shape)
            else:
                expected_value_shape = (1,)
                if isinstance(value, list):
                    value = np.array(value)
                if (
                    isinstance(value, np.ndarray)
                    and value.shape != expected_value_shape
                ):
                    raise ValueShapeError(expected_value_shape, value.shape)
        return value

    def _resize_shape(self, tensor: zarr.Array, size: int) -> None:
        """append first dimension of single array"""
        shape = list(tensor.shape)
        shape[0] = size
        tensor.resize(*shape)

    def resize_shape(self, size: int) -> None:
        """append shape of storage and dynamic tensors"""
        self.shape = (size,) + self.shape[1:]
        self.max_shape = (size,) + self.max_shape[1:]
        self._resize_shape(self._storage_tensor, size)

        if self._dynamic_tensor:
            self._resize_shape(self._dynamic_tensor, size)

        self.fs_map[".hub.dynamic_tensor"] = bytes(
            json.dumps({"shape": self.shape}), "utf-8"
        )

    def get_shape_samples(self, samples):
        """Gets full shape of dynamic_tensor(s)"""
        if isinstance(samples, int):
            shape, shape_offset = [], 0
            for i in range(1, len(self.shape)):
                if self.shape[i] is not None:
                    current = self.shape[i]
                else:
                    current = self._dynamic_tensor[samples][shape_offset]
                    shape_offset += 1
                shape.append(current)
            return np.array(shape)
        elif isinstance(samples, slice):
            shapes = self._dynamic_tensor[samples]
            for i in range(1, len(self.shape)):
                if self.shape[i] is not None:
                    shapes = np.insert(shapes, i - 1, self.shape[i], axis=1)
            return shapes
        elif isinstance(samples, list):
            shapes = np.array([self._dynamic_tensor[index] for index in samples])
            for i in range(1, len(self.shape)):
                if self.shape[i] is not None:
                    shapes = np.insert(shapes, i - 1, self.shape[i], axis=1)
            return shapes

    def combine_shape(self, shape, slice_):
        """Combines given shape with slice to get final shape"""
        if len(slice_) > shape.shape[-1]:
            raise ValueError("Slice can't be longer than shape")
        if shape.ndim == 1:  # single shape accessed
            new_shape = np.ones((0))
            for i in range(shape.shape[-1]):
                if i < len(slice_) and isinstance(slice_[i], slice):
                    start = slice_[i].start if slice_[i].start is not None else 0
                    stop = slice_[i].stop if slice_[i].stop is not None else shape[i]
                    sl = stop - start if stop != 0 else 0
                    new_shape = np.append(new_shape, sl)
                elif i >= len(slice_):
                    new_shape = np.append(new_shape, shape[i])
        else:  # slice of shapes accessed
            new_shape = np.ones(
                (shape.shape[0], 0)
            )  # new shape with rows equal to number of shapes accessed
            for i in range(shape.shape[-1]):
                if i < len(slice_) and isinstance(slice_[i], slice):
                    start = slice_[i].start if slice_[i].start is not None else 0
                    stop = slice_[i].stop
                    if stop is None:
                        sh = shape[:, i : i + 1] - start
                        sh[sh < 0] = 0  # if negative in shape, replace with 0
                        new_shape = np.append(new_shape, sh, axis=1)
                    else:
                        sl = stop - start if stop != 0 else 0
                        new_shape = np.insert(
                            new_shape, new_shape.shape[1], sl, axis=1
                        )  # inserted as last column
                elif i >= len(slice_):
                    new_shape = np.append(new_shape, shape[:, i : i + 1], axis=1)

        return np.array(new_shape).astype("int")

    def get_shape(self, slice_):
        """Gets the shape of the slice from tensor"""
        if isinstance(slice_, (int, slice)):
            slice_ = [slice_]
        if self._dynamic_tensor is None:  # returns 1D np array
            return self.combine_shape(np.array(self.shape), slice_)
        elif isinstance(slice_[0], int):  # returns 1D np array
            sample_shape = self.get_shape_samples(slice_[0])
            return self.combine_shape(sample_shape, slice_[1:])
        elif isinstance(slice_[0], (slice, list)):
            sample_shapes = self.get_shape_samples(slice_[0])
            final_shapes = self.combine_shape(sample_shapes, slice_[1:])
            return final_shapes  # returns 2D np array

    def set_shape(self, slice_, value):
        """
        Set shape of the dynamic tensor given value
        """
        if not self._enabled_dynamicness:
            return

        new_shape = self.get_shape_from_value(slice_, value)
        self.set_dynamic_shape(slice_, new_shape)

    def set_dynamic_shape(self, slice_, shape):
        """
        Set shape from the shape directly
        """
        self._dynamic_tensor[slice_[0]] = shape

    def get_shape_from_value(self, slice_, value):
        """
        create shape for multiple elements
        """
        if isinstance(slice_[0], int):
            new_shapes = self.create_shape(slice_, value)
            new_shapes = np.maximum(self._dynamic_tensor[slice_[0]], new_shapes)
        else:
            start = slice_[0].start if slice_[0].start is not None else 0
            stop = (
                slice_[0].stop if slice_[0].stop is not None else start + value.shape[0]
            )
            dt = self._dynamic_tensor[slice_[0]]
            new_shapes = []
            for i in range(start, stop):
                new_shape = self.create_shape([i] + slice_[1:], value[i - start])
                new_shape = np.maximum(dt[i - start], new_shape)
                new_shapes.append(new_shape)
        return new_shapes

    def create_shape(self, slice_, value):
        assert isinstance(slice_[0], int)
        new_shape = []
        shape_offset = 0

        value_shape = (
            list(value.shape)
            if hasattr(value, "shape") and len(list(value.shape)) > 0
            else [1]
        )

        for i in range(1, len(self.shape)):

            if self.shape[i] is None:
                if i < len(slice_):
                    if isinstance(slice_[i], slice):
                        sl = slice_[i].stop
                        shape_offset += 1
                    else:
                        sl = slice_[i] + 1
                    new_shape.append(sl)
                else:
                    new_shape.append(value_shape[shape_offset])
                    shape_offset += 1
            elif i >= len(slice_) or isinstance(slice_[i], slice):
                shape_offset += 1
        new_appended_shape = list(self.shape)
        for i, dim in enumerate(self._dynamic_dims):
            new_appended_shape[dim] = new_shape[i]
        self._delete_chunks_after_reshape(slice_[0], new_appended_shape[1:])
        return new_shape

    def _get_slice(self, slice_, real_shapes):
        # Makes slice_ which is uses relative indices (ex [:-5]) into precise slice_ (ex [10:40])
        slice_ = list(slice_)
        if real_shapes is not None:
            for r, i in enumerate(self._dynamic_dims):
                if isinstance(slice_[i], int) and slice_[i] < 0:
                    slice_[i] += real_shapes[r]
                elif isinstance(slice_[i], slice) and (
                    slice_[i].stop is None or slice_[i].stop < 0
                ):
                    slice_[i] = slice_stop_changed(
                        slice_[i], (slice_[i].stop or 0) + real_shapes[r]
                    )
        return tuple(slice_)

    def _delete_chunks_after_reshape(self, samples, new_shape: np.ndarray):
        """For a single sample or slice of samples deletes all chunks that exist out of new_shape bounds
        New shape does not include first (sample) dimension. It assumes that each sample gets the same new_shape shape
        NOTE: There is an assumption that dynamic_tensor chunks is either (1, A, B, C, ...) or (X, Infinity, Infinity, Infinity, ...)
        """
        if self.chunks[0] > 1:
            return

        if isinstance(samples, slice):
            samples_shape = self.get_shape([samples])
            for sample in range(slice.start, slice.stop, slice.step):
                self._delete_chunks_after_reshape_single_sample(
                    sample, samples_shape[sample], new_shape
                )
        else:
            assert isinstance(samples, int)
            self._delete_chunks_after_reshape_single_sample(
                samples, self.get_shape([samples]), new_shape
            )

    def _delete_chunks_after_reshape_single_sample(
        self, sample, sample_shape, new_shape
    ):
        if (sample_shape <= new_shape).all():
            return

        shapes = sample_shape
        assert len(shapes.shape) + 1 == len(self.shape)
        chunks = self._storage_tensor.chunks[1:]

        div = np.ceil(shapes / chunks).astype("int32")
        for index in np.ndindex(*div.tolist()):
            if (np.array(index) * chunks >= new_shape).any():
                try:
                    del self[".".join((sample,) + index)]
                except KeyError:
                    pass

    @property
    def chunksize(self):
        """
        Get chunk shape of the array
        """
        return self._storage_tensor.chunks

    def _get_chunking_dim(self):
        for i, d in enumerate(self.chunksize):
            if d != 1:
                return i, self.shape[i], self.chunksize[i]
        return 0, self.shape[0], self.chunksize[0]

    def commit(self):
        """ Deprecated alias to flush()"""
        self.flush()

    def flush(self):
        self._storage_tensor.store.flush()
        if self._dynamic_tensor:
            self._dynamic_tensor.store.flush()

    def close(self):
        self._storage_tensor.store.close()
        if self._dynamic_tensor:
            self._dynamic_tensor.store.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()

    @property
    def is_dynamic(self):
        return False if self._dynamic_tensor is None else True

    def disable_dynamicness(self):
        self._enabled_dynamicness = False

    def enable_dynamicness(self):
        self._enabled_dynamicness = True


def get_dynamic_dims(shape):
    return [i for i, s in enumerate(shape) if s is None]


def slice_stop_changed(slice_, new_stop):
    return slice(slice_.start, new_stop, slice_.step)
